# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Charlie Vanaret
#                 Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
# from Tedford2010
# Sellar analytical problem, p9
r"""
Sellar MDO problem
******************

The **sellar** module implements all :class:`.MDODiscipline`
included in the Sellar problem:

.. math::

   \begin{aligned}
   \text{minimize the objective function }&obj=x_{local}^2 + x_{shared,1}
   +y_0+e^{-y_1} \\
   \text{with respect to the design variables }&x_{shared},\,x_{local} \\
   \text{subject to the general constraints }
   & c_1 \leq 0\\
   & c_2 \leq 0\\
   \text{subject to the bound constraints }
   & -10 \leq x_{shared,0} \leq 10\\
   & 0 \leq x_{shared,1} \leq 10\\
   & 0 \leq x_{local} \leq 10.
   \end{aligned}

where the coupling variables are

.. math::

    \text{Discipline 0: } y_0 = x_{shared,0}^2 + x_{shared,1} +
     x_{local} - 0.2\,y_1,

and

.. math::

    \text{Discipline 1: }y_1 = \sqrt{y_0} + x_{shared,0} + x_{shared,1}.

and where the general constraints are

.. math::

   c_0 = 1-y_0/{3.16}

   c_1 = y_1/{24.} - 1
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import super
from cmath import exp, sqrt

from future import standard_library
from numpy import array, atleast_2d, complex128, ones, zeros

from gemseo.core.discipline import MDODiscipline

standard_library.install_aliases()


def get_inputs(names=None):
    """Generate initial solution

    :param names: input names (default: None)
    :type names: list(str)
    :returns: input values
    :rtype: dict
    """
    inputs = {
        "x_local": array([0.0], dtype=complex128),
        "x_shared": array([1.0, 0.0], dtype=complex128),
        "y_0": ones((1), dtype=complex128),
        "y_1": ones((1), dtype=complex128),
    }
    if names is None:
        return inputs
    return {k: inputs[k] for k in names}


class SellarSystem(MDODiscipline):
    """**SellarSystem** is the :class:`.MDODiscipline`
    implementing the computation of the Sellar's objective
    and constraints discipline."""

    def __init__(self):
        """Constructor"""
        super(SellarSystem, self).__init__(auto_detect_grammar_files=True)
        self.default_inputs = get_inputs()
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def _run(self):
        """Defines the execution of the process, given that data
        has been checked.
        Compute the outputs (= objective value and constraints at system level)
        of the Sellar analytical problem.
        """
        x_local, x_shared, y_0, y_1 = self.get_inputs_by_name(
            ["x_local", "x_shared", "y_0", "y_1"]
        )
        obj = array([self.obj(x_local, x_shared, y_0, y_1)], dtype=complex128)
        c_1 = array([self.c_1(y_0)], dtype=complex128)
        c_2 = array([self.c_2(y_1)], dtype=complex128)
        self.store_local_data(obj=obj, c_1=c_1, c_2=c_2)

    @staticmethod
    def obj(x_local, x_shared, y_0, y_1):
        """Objective function

        :param x_local: local design variables
        :type x_local: ndarray
        :param x_shared: shared design variables
        :type x_shared: ndarray
        :param y_0: coupling variable from discipline 1
        :type y_0: ndarray
        :param y_1: coupling variable from discipline 2
        :type y_1: ndarray
        :returns: Objective value
        :rtype: float
        """
        return x_local[0] ** 2 + x_shared[1] + y_0[0] ** 2 + exp(-y_1[0])

    @staticmethod
    def c_1(y_0):
        """First constraint on system level

        :param y_0: coupling variable from discipline 1
        :type y_0: ndarray
        :returns: Value of the constraint
        :rtype: float
        """
        return 3.16 - y_0[0] ** 2

    @staticmethod
    def c_2(y_1):
        """Second constraint on system level

        :param y_1: coupling variable from discipline 2
        :type y_1: ndarray
        :returns: Value of the constraint
        :rtype: float
        """
        return y_1[0] - 24.0

    def _compute_jacobian(self, inputs=None, outputs=None):
        """
        Computes the jacobian

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        # Initialize all matrices to zeros
        self._init_jacobian(inputs, outputs, with_zeros=True)
        x_local, _, y_0, y_1 = self.get_inputs_by_name(
            ["x_local", "x_shared", "y_0", "y_1"]
        )
        self.jac["c_1"]["y_0"] = atleast_2d(array([-2.0 * y_0]))
        self.jac["c_2"]["y_1"] = ones((1, 1))
        self.jac["obj"]["x_local"] = atleast_2d(array([2.0 * x_local[0]]))
        self.jac["obj"]["x_shared"] = atleast_2d(array([0.0, 1.0]))
        self.jac["obj"]["y_0"] = atleast_2d(array([2.0 * y_0[0]]))
        self.jac["obj"]["y_1"] = atleast_2d(array([-exp(-y_1[0])]))


class Sellar1(MDODiscipline):
    """**Sellar1** is the :class:`.MDODiscipline`
    implementing the 1st set of equations: y_0.
    """

    def __init__(self, residual_form=False):
        """
        Constructor

        :param residual_form: if True only residuals are computed, no Ys
        :type residual_form: bool
        """
        self.residual_form = residual_form
        super(Sellar1, self).__init__(auto_detect_grammar_files=True)
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY
        if residual_form:
            self.residual_variables = ["r_0"]
            self.output_grammar.remove_item("y_0")

        else:
            self.output_grammar.remove_item("r_0")
            self.input_grammar.remove_item("y_0")

        self.default_inputs = get_inputs(self.input_grammar.get_data_names())

    def get_attributes_to_serialize(self):
        """Defines the attributes to be serialized
        Can be overloaded by disciplines

        :returns: the list of attributes names
        :rtype: list(str)
        """
        base_d = super(Sellar1, self).get_attributes_to_serialize()
        base_d.append("residual_form")
        return base_d

    def _run(self):
        """Defines the execution of the process, given that
        data has been checked.
        Solve a coupling equation in functional form and
        compute coupling variable y_0.
        """
        x_local, x_shared, y_1 = self.get_inputs_by_name(["x_local", "x_shared", "y_1"])

        if self.residual_form:
            y_0 = self.get_inputs_by_name("y_0")

            # residual form
            r_0_out = array(
                [self.compute_r_0(x_local, x_shared, y_0, y_1)], dtype=complex128
            )
            # store outputs
            self.store_local_data(r_0=r_0_out)
        else:
            # functional form
            y_0_out = array(
                [self.compute_y_0(x_local, x_shared, y_1)], dtype=complex128
            )
            self.store_local_data(y_0=y_0_out)

    @staticmethod
    def compute_y_0(x_local, x_shared, y_1):
        """Solve the first coupling equation in functional form.

        :param x_local: vector of design variables local to discipline 1
        :type x_local: ndarray
        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :param y_1: coupling variable of discipline 2
        :type y_1: ndarray
        :returns: coupling variable y_0 of discipline 1
        :rtype: float
        """
        return sqrt(x_shared[0] ** 2 + x_shared[1] + x_local[0] - 0.2 * y_1[0])

    @staticmethod
    def compute_r_0(x_local, x_shared, y_0, y_1):
        """Evaluate the first coupling equation in residual form.

        :param x_local: vector of design variables local to discipline 1
        :type x_local: ndarray
        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :param y_0: coupling variable of discipline 1
        :type y_0: ndarray
        :param y_1: coupling variable of discipline 2
        :type y_1: ndarray
        :returns: coupling variable y_0
        :rtype: float
        """
        return Sellar1.compute_y_0(x_local, x_shared, y_1) - y_0[0]

    def _compute_jacobian(self, inputs=None, outputs=None):
        """

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        # Initialize all matrices to zeros
        self._init_jacobian(inputs, outputs, with_zeros=True)
        x_local, x_shared, y_1 = self.get_inputs_by_name(["x_local", "x_shared", "y_1"])

        inv_denom = 1.0 / (self.compute_y_0(x_local, x_shared, y_1))
        self.jac["y_0"] = {}
        self.jac["y_0"]["x_local"] = atleast_2d(array([0.5 * inv_denom]))
        self.jac["y_0"]["x_shared"] = atleast_2d(
            array([x_shared[0] * inv_denom, 0.5 * inv_denom])
        )
        self.jac["y_0"]["y_1"] = atleast_2d(array([-0.1 * inv_denom]))

        if self.residual_form:
            self.jac["r_0"]["x_local"] = atleast_2d(self.jac["y_0"]["x_local"])
            self.jac["r_0"]["x_shared"] = atleast_2d(self.jac["y_0"]["x_shared"])
            self.jac["r_0"]["y_1"] = atleast_2d(self.jac["y_0"]["y_1"])
            self.jac["r_0"]["y_0"] = -ones((1, 1))
            del self.jac["y_0"]


class Sellar2(MDODiscipline):
    """**Sellar1** is the :class:`.MDODiscipline`
    implementing the 2nd set of equations: y_1
    """

    def __init__(self, residual_form=False):
        """
        Constructor

        :param residual_form: if True only residuals are computed, no Ys
        :type residual_form: bool
        """
        self.residual_form = residual_form
        super(Sellar2, self).__init__(auto_detect_grammar_files=True)
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY
        if residual_form:
            self.residual_variables = ["r_1"]
            self.output_grammar.remove_item("y_1")
        else:
            self.output_grammar.remove_item("r_1")
            self.input_grammar.remove_item("y_1")
        self.default_inputs = get_inputs(self.input_grammar.get_data_names())

    def get_attributes_to_serialize(self):
        """Defines the attributes to be serialized
        Can be overloaded by disciplines

        :returns: the list of attributes names
        :rtype: list(str)
        """
        base_d = super(Sellar2, self).get_attributes_to_serialize()
        base_d.append("residual_form")
        return base_d

    def _run(self):
        """Defines the execution of the process, given
        that data has been checked.
        Solve a coupling equation in functional form and compute coupling
        variable y1.
        """
        x_shared, y_0 = self.get_inputs_by_name(["x_shared", "y_0"])

        if self.residual_form:
            y_1 = self.get_inputs_by_name("y_1")
            # residual form
            r_1_out = array([self.compute_r_1(x_shared, y_0, y_1)], dtype=complex128)
            # store outputs
            self.store_local_data(r_1=r_1_out)
        else:
            # functional form
            y_1_out = array([self.compute_y1(x_shared, y_0)], dtype=complex128)
            self.store_local_data(y_1=y_1_out)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """
        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        # Initialize all matrices to zeros
        self._init_jacobian(inputs, outputs, with_zeros=True)
        y_0 = self.get_inputs_by_name("y_0")
        self.jac["y_1"] = {}
        self.jac["y_1"]["x_local"] = zeros((1, 1))
        self.jac["y_1"]["x_shared"] = ones((1, 2))
        if y_0[0] < 0.0:
            self.jac["y_1"]["y_0"] = -ones((1, 1))
        elif y_0[0] == 0.0:
            self.jac["y_1"]["y_0"] = zeros((1, 1))
        else:
            self.jac["y_1"]["y_0"] = ones((1, 1))

        if self.residual_form:
            self.jac["r_1"]["x_local"] = self.jac["y_1"]["x_local"]
            self.jac["r_1"]["x_shared"] = self.jac["y_1"]["x_shared"]
            self.jac["r_1"]["y_0"] = self.jac["y_1"]["y_0"]
            self.jac["r_1"]["y_1"] = -ones((1, 1))
            del self.jac["y_1"]

    @staticmethod
    def compute_y1(x_shared, y_0):
        """Solve the second coupling equation in functional form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :param y_0: coupling variable of discipline 1
        :type y_0: ndarray
        :returns: coupling variable y_1
        :rtype: float
        """
        out = x_shared[0] + x_shared[1]
        if y_0[0].real == 0:
            y_1 = out
        elif y_0[0].real > 0:
            y_1 = y_0[0] + out
        else:
            y_1 = -y_0[0] + out
        return y_1

    @staticmethod
    def compute_r_1(x_shared, y_0, y_1):
        """Evaluate the second coupling equation in residual form.

        :param x_shared: vector of shared design variables
        :type x_shared: ndarray
        :param y_0: coupling variable of discipline 1
        :type y_0: ndarray
        :param y_1: coupling variable of discipline 2
        :type y_1: ndarray
        :returns: coupling variable y_0
        :rtype: float
        """
        return Sellar2.compute_y1(x_shared, y_0) - y_1[0]


# #=========================================================================
# # Analytical Jacobians of f and residuals
#
#
# class Jacobian(object):
#     """
#     Compute the Jacobians of objective function and residuals with respect
#     to design and coupling variables
#     """
#
#     @staticmethod
#     def jacobian_f_x(x_shared):
#         """
#         Jacobian matrix of objective function wrt design variables
#         :param x_shared: vector of design variables
#         :type x_shared: ndarray
#         :returns: Jacobian matrix
#         """
#         return array([0, 2 * x_shared[1], 1])
#
#     @staticmethod
#     def jacobian_f_y(Y):
#         """
#         Jacobian matrix of objective function wrt coupling variables
#         :param Y: vector of coupling variables
#         :type Y: ndarray
#         :returns: Jacobian matrix
#         """
#         return array([2 * Y[0], -exp(-Y[1])])
#
#     @staticmethod
#     def jacobian_residuals_x(x_shared):
#         """
#         Jacobian matrix of residual vector wrt design variables
#         :param x_shared: vector of design variables
#         :type x_shared: ndarray
#         :returns: Jacobian matrix
#         """
#         return array([[2 * x_shared[0], 1, 1], [1, 0, 1]])
#
#     @staticmethod
#     def jacobian_residuals_y(Y):
#         """
#         Jacobian matrix of residual vector wrt coupling variables
#         :param Y: vector of coupling variables
#         :type Y: ndarray
#         :returns: Jacobian matrix
#         """
#         return array([[-2 * Y[0], -0.2], [Y[0] / abs(Y[0]), -1.]])
#
#     @staticmethod
#     def jacobian_g_x():
#         """
#         Jacobian matrix of system-level constraints wrt design variables
#         :returns: Jacobian matrix
#         """
#         return array([[0., 0., 0.], [0., 0., 0.]])
#
#     @staticmethod
#     def jacobian_g_y(Y):
#         """
#         Jacobian matrix of system-level constraints wrt coupling variables
#         :param Y: vector of coupling variables
#         :type Y: ndarray
#         :returns: Jacobian matrix
#         """
#         return array([[2 * Y[0], 0], [0, 1]])



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Damien Guenot - 26 avr. 2016
#       :author: Francois Gallard, refactoring
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Driver library
==============

A driver library aims to solve an :class:`.DriverLib`
using a particular algorithm from a particular family of numerical methods.
This algorithm will be in charge of evaluating the objective and constraints
functions at different points of the design space, using the
:meth:`.DriverLib.execute` method.
The most famous kinds of numerical methods to solve an optimization problem
are optimization algorithms and design of experiments (DOE). A DOE driver
browses the design space agnostically, i.e. without taking into
account the function evaluations. On the contrary, an optimization algorithm
uses this information to make the journey through design space
as relevant as possible in order to reach as soon as possible the optimum.
These families are implemented in :class:`.DOELibrary`
and :class:`.OptimizationLibrary`.
"""
from __future__ import absolute_import, division, unicode_literals

import inspect
from builtins import str
from os.path import dirname, exists, join

from future import standard_library
from numpy import ones_like, where, zeros_like

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.algos.opt_result import OptimizationResult
from gemseo.algos.stop_criteria import (
    DesvarIsNan,
    FunctionIsNan,
    MaxIterReachedException,
    TerminationCriterion,
)
from gemseo.core.grammar import InvalidDataException
from gemseo.core.json_grammar import JSONGrammar
from gemseo.third_party.tqdm import Tqdm
from gemseo.utils.source_parsing import SourceParsing

standard_library.install_aliases()


from gemseo import LOGGER


class DriverLib(object):
    """Abstract class for DOE & optimization libraries interfaces

    Lists available methods in the library for the proposed
    problem to be solved

    To integrate an optimization package, inherit from this class
    and put your file in gemseo.algos.doe or gemseo.algo.opt packages


    """

    USER_DEFINED_GRADIENT = OptimizationProblem.USER_GRAD
    COMPLEX_STEP_METHOD = OptimizationProblem.COMPLEX_STEP
    FINITE_DIFF_METHOD = OptimizationProblem.FINITE_DIFFERENCES

    DIFFERENTIATION_METHODS = [
        USER_DEFINED_GRADIENT,
        COMPLEX_STEP_METHOD,
        FINITE_DIFF_METHOD,
    ]

    LIB = "lib"
    INTERNAL_NAME = "internal_algo_name"
    OPTIONS_DIR = "options"
    OPTIONS_MAP = {}
    REQUIRE_GRAD = "require_grad"
    HANDLE_EQ_CONS = "handle_equality_constraints"
    HANDLE_INEQ_CONS = "handle_inequality_constraints"
    POSITIVE_CONSTRAINTS = "positive_constraints"
    INEQ_TOLERANCE = "ineq_tolerance"
    EQ_TOLERANCE = "eq_tolerance"
    USE_DATABASE_OPTION = "use_database"
    NORMALIZE_DESIGN_SPACE_OPTION = "normalize_design_space"
    WEBSITE = "website"
    DESCRIPTION = "description"
    MAX_DS_SIZE_PRINT = 40

    def __init__(self):
        """
        Constructor
        """
        # Library settings and check
        self.lib_dict = {}
        self.algo_name = None
        self.internal_algo_name = None
        self.problem = None
        self.opt_grammar = None
        self._progress_bar = None

    def init_options_grammar(self, algo_name):
        """Initializes the options grammar

        :param algo_name: name of the algorithm
        """
        # Try algo name convention which has the prioryty over
        # library options
        basename = algo_name.upper() + "_options.json"
        lib_dir = inspect.getfile(self.__class__)
        opt_dir = join(dirname(lib_dir), self.OPTIONS_DIR)
        algo_schema_file = join(opt_dir, basename)
        if exists(algo_schema_file):
            schema_file = algo_schema_file
        else:
            # Try to load library options convention by default
            basename = self.__class__.__name__.upper() + "_options.json"
            lib_schema_file = join(opt_dir, basename)

            if exists(lib_schema_file):
                schema_file = lib_schema_file
            else:
                msg = "Options grammar file " + algo_schema_file
                msg += " for algorithm: " + algo_name
                msg += " not found. And library options grammar file "
                msg += lib_schema_file + " not found either."
                raise ValueError(msg)

        self.opt_grammar = JSONGrammar(
            algo_name, schema_file=schema_file, grammar_type="input"
        )

        descr_dict = SourceParsing.get_options_doc(self.__class__._get_options)
        self.opt_grammar.add_description(descr_dict)
        return self.opt_grammar

    def init_progress_bar(self, max_iter, message):
        """Initializes the progress bar

        :param max_iter: maximum number of calls
        :param message: message to display at the beginning
        """
        if max_iter <= 0:
            raise ValueError("max_iter must be >=1, got " + str(max_iter))
        self._progress_bar = Tqdm(max_iter, desc=message, mininterval=0.1)

    def __get_signed_obj_val(self):
        """
        Gets the objective value with the right sign
        """
        value = self.problem.objective.last_eval
        # if maximization problem, take the opposite
        if value is not None and not self.problem.minimize_objective:
            value = -value
        return value

    @property
    def algorithms(self):
        """Return the available algorithms."""
        return list(self.lib_dict.keys())

    def close_progress_bar(self):
        """Closes the progress bar"""
        if self._progress_bar is not None:
            value = self.__get_signed_obj_val()
            self._progress_bar.close(value)

    def iterate_progress_bar(self):
        """Iterates the progress bar"""
        try:
            value = self.__get_signed_obj_val()
            self._progress_bar.next(value)
        except StopIteration:
            raise MaxIterReachedException

    def _pre_run(self, problem, algo_name, **options):
        """To be overriden by subclasses
        Specific method to be executed just before _run method call

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
        :param options: the options dict for the algorithm,
            see associated JSON file
        """

    def _post_run(self, problem, algo_name, result, **options):  # pylint: disable=W0613
        """To be overriden by subclasses
        Specific method to be executed just after _run method call

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
        :param result: result of the run such as an OptimizationResult
        :param options: the options dict for the algorithm,
            see associated JSON file
        """
        result.log_me()
        problem.solution = result
        if result.x_opt is not None:
            problem.design_space.set_current_x(result)
        if problem.design_space.dimension <= self.MAX_DS_SIZE_PRINT:
            problem.design_space.log_me()

    def _process_options(self, **options):
        """After _get_options is called, the options are
        converted to algorithm specific options, and checked

        :param options: driver options
        """

        for option_key in list(options.keys()):  # Copy keys on purpose
            # Remove extra options
            if not self.opt_grammar.is_data_name_existing(option_key):
                del options[option_key]
            elif option_key == self.INEQ_TOLERANCE:
                self.problem.ineq_tolerance = options[option_key]
                del options[option_key]
            elif option_key == self.EQ_TOLERANCE:
                self.problem.eq_tolerance = options[option_key]
                del options[option_key]
            elif options[option_key] is None:
                del options[option_key]

        try:
            self.opt_grammar.load_data(options)
        except InvalidDataException:
            raise ValueError("Invalid options for algorithm " + self.opt_grammar.name)
        for option_key in list(options.keys()):  # Copy keys on purpose
            lib_option_key = self.OPTIONS_MAP.get(option_key)
            # Overload with specific keys
            if lib_option_key is not None:
                options[lib_option_key] = options[option_key]
                if lib_option_key != option_key:
                    del options[option_key]

        return options

    def execute(self, problem, algo_name=None, **options):
        """Executes the driver

        :param problem: the problem to be solved
        :param algo_name: name of the algorithm
            if None, use self.algo_name
            which may have been set by the factory (Default value = None)
        :param options: the options dict for the algorithm
        """
        self.problem = problem
        if algo_name is not None:
            self.algo_name = algo_name
        if self.algo_name is None:
            raise ValueError(
                "Algorithm name must be either passed as "
                + "argument or set by the attribute self.algo_name"
            )
        self.__check_algorithm(self.algo_name)
        self.init_options_grammar(self.algo_name)
        use_database = options.get(self.USE_DATABASE_OPTION, True)
        normalize = options.get(self.NORMALIZE_DESIGN_SPACE_OPTION, True)
        options = self._get_options(**options)
        self.internal_algo_name = self.lib_dict[self.algo_name][self.INTERNAL_NAME]
        problem.check()
        problem.preprocess_functions(
            normalize=normalize, use_database=use_database, round_ints=True
        )
        try:
            self._pre_run(problem, self.algo_name, **options)
            result = self._run(**options)
        except TerminationCriterion as error:
            result = self._termination_criterion_raised(error)
        finally:
            self.close_progress_bar()
            self.problem.clear_listeners()

        self._post_run(problem, algo_name, result, **options)

        return result

    def _termination_criterion_raised(self, error):  # pylint: disable=W0613
        """Retrieve the best known iterate when max iter has been reached

        :param error: the obtained error from the algorithm
        """
        if len(error.args) >= 1:
            message = error.args[0]
        else:
            message = ""
        if isinstance(error, MaxIterReachedException):
            message = "Maximum number of iterations reached, "
            message += "the solver has been stopped."
        elif isinstance(error, FunctionIsNan):
            message = "Function value or gradient or constraint is NaN, "
            message += "and problem.stop_if_nan is set to True. "
            message += "The solver has been stopped."
        elif isinstance(error, DesvarIsNan):
            message = "Design variables are NaN, "
            message += "The solver has been stopped."

        result = self.get_optimum_from_database(message)
        return result

    def get_optimum_from_database(self, message=None, status=None):
        """Retrieves the optimum from the database and
        builds an optimization result object from it

        :param message: Default value = None)
        :param status: Default value = None)

        """
        problem = self.problem
        x_0 = problem.database.get_x_by_iter(0)
        # compute the best feasible or infeasible point
        f_opt, x_opt, is_feas, c_opt, c_opt_grad = problem.get_optimum()
        if f_opt is not None and not problem.minimize_objective:
            f_opt = -f_opt
        return OptimizationResult(
            x_0=x_0,
            x_opt=x_opt,
            f_opt=f_opt,
            optimizer_name=self.algo_name,
            message=message,
            status=status,
            n_obj_call=problem.objective.n_calls,
            is_feasible=is_feas,
            constraints_values=c_opt,
            constraints_grad=c_opt_grad,
        )

    def _get_options(self, **options):
        """Retrieves the options of the library
        To be overloaded by subclasses
        Used to define default values for options using keyword arguments

        :param options: options of the driver
        """
        raise NotImplementedError()

    def _run(self, **options):
        """Runs the algorithm, to be overloaded by subclasses

        :param options: the options dict for the algorithm
        """
        raise NotImplementedError()

    def __check_algorithm(self, algo_name):
        """Check that algorithm required by user is available.
        Set optimization library
        Set algorithm name according to optimization library requirements
        :param algo_name : name of algorithm
        """
        if algo_name not in self.lib_dict:
            raise KeyError(
                "Requested optimization algorithm"
                + " %s is not in list of available algorithms %s"
                % (algo_name, list(self.lib_dict.keys()))
            )

    @staticmethod
    def _display_result(result):
        """Displays the optimization result

        :param result: the result to display
        """
        LOGGER.info("Algorithm execution finished, result is:")
        LOGGER.info(result)

    @staticmethod
    def is_algorithm_suited(algo_dict, problem):
        """Checks if the algorithm is suited to the problem
        according to its algo dict

        :param algo_dict: the algorithm characteristics
        :param problem: the opt_problem to be solved
        """
        raise NotImplementedError()

    def filter_adapted_algorithms(self, problem):
        """Filters the algorithms capable of solving the problem

        :param problem: the opt_problem to be solved
        :returns: the list of adapted algorithms names
        """
        available = []
        for algo_name, algo_dict in self.lib_dict.items():
            if self.is_algorithm_suited(algo_dict, problem):
                available.append(algo_name)
        return available

    def is_algo_requires_grad(self, algo_name):
        """
        Returns True if the algorithm requries a gradient evaluation

        :param algo_name: name of the algorithm
        """
        lib_alg = self.lib_dict.get(algo_name, None)
        if lib_alg is None:
            raise ValueError("Algorithm " + str(algo_name) + " is not available !")

        return lib_alg.get(self.REQUIRE_GRAD, False)

    def get_x0_and_bounds_vects(self, normalize_ds):
        """
        Gets x0, bounds, normalized or not depending
        on algo options, all as numpy arrays

        :param normalize_ds: if True, normalizes all input vars
               that are not integers, according to design space
               normalization policy
        :returns: x, lower bounds, upper bounds
        """
        dspace = self.problem.design_space
        l_b = dspace.get_lower_bounds()
        u_b = dspace.get_upper_bounds()

        # remove normalization from options for algo
        if normalize_ds:
            norm_dict = dspace.normalize
            norm_array = dspace.dict_to_array(norm_dict)
            xvec = self.problem.get_x0_normalized()
            l_b = where(norm_array, zeros_like(xvec), l_b)
            u_b = where(norm_array, ones_like(xvec), u_b)
        else:
            xvec = self.problem.design_space.get_current_x()

        return xvec, l_b, u_b



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Dataset
=======

The :mod:`~gemseo.core.dataset` module implements the concept of dataset
which is a key element for machine learning, post-processing,
data analysis, ...

A :class:`.Dataset` is an object
defined by data stored as a dictionary of 2D numpy arrays,
whose rows are samples, a.k.a. realizations, and columns are features,
a.k.a. parameters or variables. The indices of this dictionary are either
names of groups of variables or names of variables.
A :class:`.Dataset` is also defined by
a list of variables names, a dictionary of variables sizes
and a dictionary of variables groups.

A :class:`.Dataset` can be set either from a numpy array or a file.
An :class:`.AbstractFullCache` or an :class:`.OptimizationProblem`
can also be exported to a :class:`.Dataset`
using :meth:`.AbstractFullCache.export_to_dataset`
and :meth:`.OptimizationProblem.export_to_dataset` respectively.

From a :class:`.Dataset`, we can easily access its length and get the data,
either as 2D array or as dictionaries indexed by the variables names.
We can get either the whole data,
data associated to a group or data associated to a list of variables.
It is also possible to export the :class:`.Dataset`
to an :class:`.AbstractFullCache` or a pandas DataFrame.

"""
from __future__ import absolute_import, division, unicode_literals

from numbers import Number

from future import standard_library
from numpy import array, concatenate, genfromtxt, hstack, ndarray, unique
from six import string_types

from gemseo.caches.cache_factory import CacheFactory
from gemseo.post.dataset.factory import DatasetPlotFactory
from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.py23_compat import _long

standard_library.install_aliases()

from gemseo import LOGGER


class Dataset(object):
    """A Dataset is an object defined by data stored as a 2D numpy array,
    whose rows are samples, a.k.a. realizations, and columns are properties,
    a.k.a. parameters, variables or features. A dataset is also defined by
    a list of variables names, a dictionary of variables sizes
    and a dictionary of variables types. We can easily access its length
    and get the data, either as a 2D array or as a list of dictionaries
    indexed by the variables names. It is also possible to export the dataset
    to a :class:`.AbstractFullCache` or a pandas DataFrame."""

    PARAMETER_GROUP = "parameters"
    DESIGN_GROUP = "design_parameters"
    FUNCTION_GROUP = "functions"
    INPUT_GROUP = "inputs"
    OUTPUT_GROUP = "outputs"
    DEFAULT_GROUP = PARAMETER_GROUP
    DEFAULT_NAMES = {
        PARAMETER_GROUP: "x",
        DESIGN_GROUP: "dp",
        FUNCTION_GROUP: "func",
        INPUT_GROUP: "in",
        OUTPUT_GROUP: "out",
    }

    HDF5_CACHE = "HDF5Cache"
    MEMORY_FULL_CACHE = "MemoryFullCache"

    def __init__(self, name=None, by_group=True):
        """Constructor.

        :param str name: dataset name.
        :param bool group: if True, store the data by group. Otherwise,
            store them by variables. Default: True
        """
        self.name = name or self.__class__.__name__
        self._names = {}  # key = group, value = varnames
        self._groups = {}  # key = varname, value = group
        self.sizes = {}  # key = varname, value = varsize
        self._positions = {}
        self.dimension = {}  # key = group, value = groupsize
        self.length = 0
        self.data = {}
        self._group = by_group
        self.strings_encoding = None
        self._cached_inputs = []
        self._cached_outputs = []
        self._plot_factory = DatasetPlotFactory()
        self.metadata = {}

    def _clean(self):
        """ Clean dataset. """
        self._names = {}
        self._groups = {}
        self.sizes = {}
        self._positions = {}
        self.dimension = {}
        self.length = 0
        self.data = {}
        self.strings_encoding = None
        self._cached_inputs = []
        self._cached_outputs = []
        self.metadata = {}

    def is_group(self, group_name):
        """Returns True is group_name is a group.

        :param str group_name: group_name
        """
        return group_name in self._names

    def is_variable(self, variable_name):
        """Returns True is variable_name is a group.

        :param str variable_name: variable_name
        """
        return variable_name in self._groups

    def is_empty(self):
        """ Returns True if the dataset is empty. """
        return self.n_samples == 0

    def get_names(self, group_name):
        """Returns names for a given group.

        :param str group_name: group_name
        """
        return self._names.get(group_name)

    def get_group(self, variable_name):
        """Returns group for a given variable name.

        :param str variable_name: variable_name
        """
        return self._groups.get(variable_name)

    @property
    def variables(self):
        """ Names of variables. """
        return sorted(self._groups.keys())

    @property
    def groups(self):
        """ Names of the groups of variables. """
        return sorted(self._names.keys())

    def __str__(self):
        """ String representation. """
        string = self.name + "\n"
        string += "| Number of samples: " + str(self.n_samples) + "\n"
        string += "| Number of variables: " + str(self.n_variables) + "\n"
        string += "| Variables names and sizes by group: \n"
        for group, varnames in self._names.items():
            varnames = [name + " (" + str(self.sizes[name]) + ")" for name in varnames]
            varnames = ", ".join(varnames) + "\n"
            if varnames != "\n":
                string += "| - " + group + ": " + varnames
        total = str(sum(self.dimension.values()))
        string += "| Number of dimensions (total = " + total + ") by group:\n"
        for group, size in self.dimension.items():
            string += "| - " + group + ": " + str(size) + "\n"
        return string

    def __check_new_variable(self, variable):
        """Raise ValueError if variable is already defined.

        :param str variable: variable name.
        """
        if self.is_variable(variable):
            raise ValueError(variable + " is already defined.")
        if not isinstance(variable, string_types):
            raise TypeError("variable name must be a string.")

    def __check_new_group(self, group):
        """Raise ValueError if group is already defined.

        :param str group: group name.
        """
        if self.is_group(group):
            raise ValueError(group + " is already defined.")
        if not isinstance(group, string_types):
            raise TypeError("Group name must be a string.")

    def __check_length_consistency(self, length):
        """Raises ValueError if the length is different from the length of
        the dataset and if the latter is different from zero.

        :param int length: length.
        """
        if self.length != 0 and length != self.length:
            raise ValueError(
                "The number of rows of data must be equal to the"
                " length of the dataset."
            )
        self.length = length

    def __check_data_consistency(self, data):
        """Raises ValueError if the data is not a 2D numpy array
        or if its length is different from the length of the dataset.

        :param array data: data.
        """
        if not isinstance(data, ndarray) or len(data.shape) != 2:
            raise TypeError("data must be a 2D numpy array.")
        self.__check_length_consistency(data.shape[0])

    @staticmethod
    def __check_variables_format(variables):
        """Raises TypeError if the format of variables is wrong.

        :param list(str) variables: list of variables names.
        """
        if not isinstance(variables, list):
            raise TypeError("variables must be a list of strings.")
        if any([not isinstance(name, string_types) for name in variables]):
            raise TypeError("variables must be a list of strings.")

    @staticmethod
    def __check_sizes_format(sizes, variables, dimension):
        """Raises TypeError if the format of sizes is wrong.

        :param dict(int) sizes: dictionary of variables sizes.
        :param list(str) variables: list of variables names.
        :param int dimension: data dimension.
        """

        def is_size(size):
            return isinstance(size, (int, _long)) and size > 0

        if not isinstance(sizes, dict):
            raise TypeError("sizes must be a dictionary of positive integers.")
        if any([not is_size(sizes.get(name)) for name in variables]):
            raise TypeError("sizes must be a dictionary of positive integers.")
        total = sum([sizes[name] for name in variables])
        if total != dimension:
            raise ValueError(
                "The sum of variables sizes ({}) must be equal "
                "to the data dimension ({}).".format(total, dimension)
            )

    def __check_variables_sizes(self, variables, sizes, dimension):
        """Raises ValueError if the sum of the variables sizes is different
        from the number of data columns.

        :param list(str) variables: list of variables names.
        :param dict sizes: variables sizes.
        :param int dimension: data dimension.
        """
        if variables is not None:
            self.__check_variables_format(variables)
            if sizes is not None:
                self.__check_sizes_format(sizes, variables, dimension)

    def __get_default_group_variables(self, group, dimension, varname=None):
        """ Returns default variables names for a given group."""
        default_name = varname or self.DEFAULT_NAMES.get(group) or group
        variables = [default_name + "_" + str(index) for index in range(dimension)]
        sizes = {name: 1 for name in variables}
        groups = {name: group for name in variables}
        return variables, sizes, groups

    def __set_group_data(self, data, group, variables, sizes):
        """ Set data for a given group. """
        if self._group:
            self.data[group] = data
        else:
            array_to_dict = DataConversion.array_to_dict
            self.data.update(array_to_dict(data, variables, sizes))

    def __set_variable_data(self, name, data, group):
        """ Set data for a given group. """
        if self._group:
            if not self.is_group(group):
                self.data[group] = data
            else:
                self.data[group] = hstack((self.data[group], data))
        else:
            self.data[name] = data

    def __set_group_properties(self, group, variables, sizes, cache_as_input):
        """ Set properties for a given group. """
        self.sizes.update(sizes)
        self._groups.update({name: group for name in variables})
        self._names[group] = variables
        self.dimension[group] = sum([sizes[name] for name in variables])
        start = 0
        for name in variables:
            self._positions[name] = [start, start + self.sizes[name] - 1]
            if self._group:
                start += self.sizes[name]
            if cache_as_input:
                self._cached_inputs.append(name)
            else:
                self._cached_outputs.append(name)

    def __set_variable_properties(self, variable, group, size, cache_as_input):
        """ Set properties for a given variable. """
        self.sizes[variable] = size
        self._groups[variable] = group
        if not self.is_group(group):
            self._names[group] = [variable]
            self.dimension[group] = self.sizes[variable]
            self._positions[variable] = [0, self.sizes[variable] - 1]
        else:
            self._names[group].append(variable)
            if self._group:
                ncols = self.dimension[group]
                self._positions[variable] = [ncols, ncols + self.sizes[variable] - 1]
            else:
                self._positions[variable] = [0, self.sizes[variable] - 1]
            self.dimension[group] += self.sizes[variable]
        if cache_as_input:
            self._cached_inputs.append(variable)
        else:
            self._cached_outputs.append(variable)

    def add_group(
        self, group, data, variables=None, sizes=None, varname=None, cache_as_input=True
    ):
        """Add variable.

        :param str group: group name.
        :param ndarray data: data.
        :param list(str) variables: list of variables names.
        :param dict sizes: dictionary of variables sizes.
        :param str varname: variable name used if variables is None.
            If None, use the default variable name for group if it exists;
            otherwise, use the group name. Default: None.
        :param bool cache_as_input: cache as input when export to cache.
            Otherwise, as output. Default: True
        """
        self.__check_new_group(group)
        self.__check_data_consistency(data)
        self.__check_variables_sizes(variables, sizes, data.shape[1])
        if variables is None or sizes is None:
            get = self.__get_default_group_variables
            variables, sizes, _ = get(group, data.shape[1], varname)
        self.__set_group_data(data, group, variables, sizes)
        self.__set_group_properties(group, variables, sizes, cache_as_input)

    def add_variable(self, name, data, group=DEFAULT_GROUP, cache_as_input=True):
        """Add variable.

        :param str name: variable name.
        :param ndarray data: data.
        :param str group: group name. Default: DEFAULT_GROUP.
        :param bool cache_as_input: cache as input when export to cache.
            Otherwise, as output. Default: True
        """
        self.__check_new_variable(name)
        self.__check_data_consistency(data)
        self.__set_variable_data(name, data, group)
        self.__set_variable_properties(name, group, data.shape[1], cache_as_input)

    def __get_strings_encoding(self, data):
        """ Returns strings encoding. """
        self.strings_encoding = {name: {} for name in self._groups}
        encoding = {}
        if str(data.dtype).startswith(("|S", "<U")):
            data, encoding = self.__force_array_to_float(data)
        return data, encoding

    def set_from_array(
        self, data, variables=None, sizes=None, groups=None, default_name=None
    ):
        """Set Dataset from a numpy array or a dictionary of arrays

        :param array data: dataset.
        :param list(str) variables: list of variables names.
        :param dict(int) sizes: list of variables sizes.
        :param dict(str) groups: list of variables groups.
        :param str default_name: default variable name.
        """
        self._clean()
        self.__check_data_consistency(data)
        if variables is None:
            group = self.DEFAULT_GROUP
            get = self.__get_default_group_variables
            variables, sizes, groups = get(group, data.shape[1], default_name)
        else:
            self.__check_variables_format(variables)
        if sizes is None:
            sizes = {name: 1 for name in variables}
        self.__check_sizes_format(sizes, variables, data.shape[1])
        if groups is None:
            groups = {name: self.DEFAULT_GROUP for name in variables}
        self.__check_groups_format(groups, variables)
        self.__set_data_properties(variables, sizes, groups)
        data, encoding = self.__get_strings_encoding(data)
        self.__set_data(data, variables, encoding)

    def __check_groups_format(self, groups, variables):
        """ Check groups format and update it if necessary"""
        if not isinstance(groups, dict):
            raise TypeError(
                "groups must be a dictionary indexed by variables"
                "names whose values are strings."
            )
        for name in variables:
            if groups.get(name) is None:
                groups.update({name: self.DEFAULT_GROUP})
            elif not isinstance(groups[name], string_types):
                raise TypeError(
                    "groups must be a dictionary indexed "
                    "by variables names whose values are strings."
                )

    def __set_data(self, data, variables, encoding):
        """ Set data. """
        indices = {group: [] for group in self._names}
        start = 0
        end = 0
        for name in variables:
            end = start + self.sizes[name] - 1
            name_indices = list(range(start, end + 1))
            start = end + 1
            indices[self._groups[name]] += name_indices
            for key in encoding:
                if key in name_indices:
                    index = name_indices.index(key)
                    self.strings_encoding[name][index] = encoding[key]
            if not self._group:
                self.data[name] = data[:, name_indices]
        if self._group:
            for group in self._names:
                self.data[group] = data[:, indices[group]]

    def _set_variables_positions(self):
        """ Set variables positions. """
        for varnames in self._names.values():
            start = 0
            for varname in varnames:
                self._positions[varname] = [start, start + self.sizes[varname] - 1]
                if self._group:
                    start += self.sizes[varname]
                else:
                    start = 0

    def __set_data_properties(self, variables, sizes, groups):
        """Set properties for the whole data.

        :param list(str) variables: list of variables names.
        :param dict(int) sizes: dictionary of variables sizes.
        :param dict(str) groups: dictionary of variables groups.
        """
        for name in variables:
            if not self.is_group(groups[name]):
                self._names[groups[name]] = [name]
            else:
                self._names[groups[name]].append(name)
            self.sizes[name] = sizes[name]
            self._groups[name] = groups[name]
        for group, names in self._names.items():
            self.dimension[group] = sum([self.sizes[name] for name in names])
            if group == self.OUTPUT_GROUP:
                self._cached_outputs += names
            else:
                self._cached_inputs += names
        self._set_variables_positions()

    @staticmethod
    def __force_array_to_float(data):
        """Force a ndarray type to float.

        :param ndarray data: dataset.
        :return: data with float type.
        :rtype: ndarray
        """

        def __is_not_float(obj):
            """Return True if an object cannot be cast to float.

            :param obj: object to test.
            """
            try:
                float(obj)
                return False
            except ValueError:
                return True

        first_row = data[0, :]
        str_indices = [
            index for index, element in enumerate(first_row) if __is_not_float(element)
        ]

        strings_encoding = {}
        for index in range(data.shape[1]):
            if index in str_indices:
                column = list(data[:, index])
                encoding = dict(enumerate(unique(column)))
                strings_encoding[index] = encoding
                data[:, index] = unique(column, return_inverse=True)[1]

        data = data.astype("float")
        return data, strings_encoding

    def set_from_file(
        self,
        filename,
        variables=None,
        sizes=None,
        groups=None,
        delimiter=",",
        header=True,
    ):
        """Set Dataset from a file.

        :param str filename: file name.
        :param list(str) variables: list of variables names.
        :param dict(int) sizes: list of variables sizes.
        :param dict(str) groups: list of variables groups.
        :param str delimiter: field delimiter.
        :param bool header: if True, read the variables names
            on the first line of the file. Default: True.
        """
        self._clean()
        data = genfromtxt(filename, delimiter=delimiter, dtype="str")
        if header:
            if variables is None:
                variables = data[0, :].tolist()
            start_read = 1
        else:
            start_read = 0
        self.set_from_array(data[start_read:, :], variables, sizes, groups)

    def set_metadata(self, name, value):
        """Set metadata attribute.
        :param string name: Metadata attribute name.
        :param value: Metadata attribute value.
        """
        self.metadata[name] = value

    def _get_columns_names(self, as_list=False, start=0):
        """Return the names of the data columns. If dim(x)=1,
        its column name is 'x' while if dim(y)=2, its columns names
        are either 'x_0' and 'x_1' or ['x',0] and ['x',1].

        :param bool as_tuple: if True, return the name as a tuple.
            if False, return the name as a string. Default: False.
        :param int start: first index for components of a variable.
            Default: 0 ('x_0', 'x_1', ...).
        """
        tmp = []
        for _, names in self._names.items():
            for name in names:
                if as_list:
                    tmp += [
                        [self._groups[name], name, str(index + start)]
                        for index in range(self.sizes[name])
                    ]
                else:
                    if self.sizes[name] == 1:
                        tmp += [name]
                    else:
                        tmp += [
                            name + "_" + str(index + start)
                            for index in range(self.sizes[name])
                        ]
        return tmp

    def get_data_by_group(self, group, as_dict=False):
        """Returns data associated with a group.

        :param str group: variable group.
        :param bool as_dict: if True, return outputs values as dictionary.
            Default: False.
        """
        if not self.is_group(group):
            raise ValueError("{} is not an available group.".format(group))
        if group in self.data:
            data = self.data[group]
            if as_dict:
                data = DataConversion.array_to_dict(
                    self.data[group], self._names[group], self.sizes
                )
        else:
            data = {name: self.data[name] for name in self._names[group]}
            if not as_dict:
                data = DataConversion.dict_to_array(data, self._names[group])
        return data

    def get_data_by_names(self, names, as_dict=True):
        """Get data by variables names.

        :param list(str): names.
        :param bool as_dict: if True, return values as dictionary.
        """
        if isinstance(names, string_types):
            names = [names]
        dict_to_array = DataConversion.dict_to_array
        if not self._group:
            data = {name: self.data.get(name) for name in names}
        else:
            data = {}

            for name in names:
                indices = list(
                    range(self._positions[name][0], self._positions[name][1] + 1)
                )
                data[name] = self.data[self._groups[name]][:, indices]
        if not as_dict:
            data = dict_to_array(data, names)
        return data

    def get_all_data(self, by_group=True, as_dict=False):
        """Returns all data.

        :param str group: variable group.
        :param bool as_dict: if True, return outputs values as dictionary.
            Default: False.
        """
        if by_group:
            data = {
                group: self.get_data_by_group(group, as_dict) for group in self._names
            }
            if not as_dict:
                data = (data, self._names, self.sizes)
        else:
            if not as_dict:
                data = concatenate(
                    tuple([self.get_data_by_group(group) for group in self._names]), 1
                )
                names = [
                    item for sublist in list(self._names.values()) for item in sublist
                ]
                data = (data, names, self.sizes)
            else:
                data = {}
                for group in self._names:
                    data.update(self.get_data_by_group(group, as_dict))
        return data

    @property
    def n_variables(self):
        """ Return the number of variables. """
        return len(self._groups)

    def n_variables_by_group(self, group):
        """Return the number of variables for a group.

        :param str group: group name.
        """
        return len(self._names[group])

    @property
    def n_samples(self):
        """ Return the number of samples. """
        return self.length

    def __len__(self):
        """ Length of the object """
        return self.length

    def __bool__(self):
        """ returns True is the dataset is not empty. """
        return not self.is_empty()

    def export_to_dataframe(self, copy=True):
        """Export dataset to Dataframe.

        :param bool copy: If True, copy data. Otherwise, use reference.
            Default: True.
        """
        from pandas import DataFrame

        row1 = []
        row2 = []
        row3 = []
        for column in self._get_columns_names(True):
            row1.append(column[0])
            row2.append(column[1])
            row3.append(column[2])
        columns = [array(row1), array(row2), array(row3)]
        data = self.get_all_data(False, False)
        return DataFrame(data[0], columns=columns, copy=copy)

    def export_to_cache(
        self,
        inputs=None,
        outputs=None,
        cache_type=MEMORY_FULL_CACHE,
        cache_hdf_file=None,
        cache_hdf_node_name=None,
        **options
    ):
        """Export dataset to cache.

        :param list(str) inputs: names of the inputs to cache. If None,
            use all inputs. Default: None.
        :param list(str) outputs: names of the outputs to cache. If None,
            use all outputs. Default: None.
        :param str cache_type: type of cache to use.
        :param str cache_hdf_file: the file to store the data,
            mandatory when HDF caching is used
        :param str cache_hdf_node_name: name of the HDF
            dataset to store the discipline
            data. If None, self.name is used
        """
        if inputs is None:
            inputs = self._cached_inputs
        if outputs is None:
            outputs = self._cached_outputs
        create_cache = CacheFactory().create
        cache_hdf_node_name = cache_hdf_node_name or self.name
        if cache_type == self.HDF5_CACHE:
            cache = create_cache(
                cache_type,
                hdf_file_path=cache_hdf_file,
                hdf_node_path=cache_hdf_node_name,
                **options
            )
        else:
            cache = create_cache(cache_type, **options)
        for sample in range(len(self)):
            in_values = {name: self[(sample, name)][name][0] for name in inputs}
            out_values = {name: self[(sample, name)][name][0] for name in outputs}
            cache.cache_outputs(in_values, inputs, out_values, outputs)
        return cache

    def plot(self, name, show=True, save=False, **options):
        """Finds the appropriate library and executes
        the post processing on the problem

        :param str name: the post processing name
        :param show: show the figure (default: True)
        :param save: save the figure (default: False)
        :param options: options for the post method, see its package
        """
        options.update({"show": show, "save": save})
        post = self._plot_factory.create(name, dataset=self)
        post.execute(**options)
        return post

    def __getitem(self, indices, names):
        """Get items associated with sample indices and variables names.

        :param list(int) indices: samples indices.
        :param list(str) names: variables names.
        """
        for index in indices:
            is_lower = index < -1
            is_greater = index >= self.length
            is_int = isinstance(index, int)
            if index == self.n_samples:
                raise IndexError
            if is_lower or is_greater or not is_int:
                raise ValueError("{} is not a sample index.".format(index))
        for name in names:
            if name not in self._groups:
                raise ValueError("{} is not a variable name.".format(name))
        item = self.get_data_by_names(names)
        if len(indices) == 1:
            indices = indices[0]
        item = {name: value[indices, :] for name, value in item.items()}
        return item

    def __getitem__(self, key):
        """Gets item, where key is either int, str, list(int), list(str),
        (int, str), (int, list(str)), (list(int), str), (int, list(str)),
        or (list(int), list(str))."""
        indices = list(range(0, self.length))
        type_error_msg = "___getitem__ uses one of these argument types: "
        type_error_msg += "int, str, "
        type_error_msg += "list(int), list(str), "
        type_error_msg += "(int, str), (int, list(str)), "
        type_error_msg += "(list(int), str), (int, list(str)) "
        type_error_msg += "or (list(int), list(str)). "

        def getitem_list(index):
            """ Gets item when index is a list."""
            if all(isinstance(elem, string_types) for elem in index):
                item = self.__getitem(indices, index)
            elif all(isinstance(elem, Number) for elem in index):
                item = self.__getitem(index, self.variables)
            else:
                raise TypeError(type_error_msg)
            return item

        def getitem_tuple(tpl):
            """ Gets item when index is a tuple."""
            if isinstance(tpl[0], Number):
                indices = [tpl[0]]
            elif isinstance(tpl[0], slice):
                indices = list(range(tpl[0].start, tpl[0].stop, tpl[0].step or 1))
            elif isinstance(tpl[0], list):
                if all(isinstance(elem, Number) for elem in tpl[0]):
                    indices = tpl[0]
                else:
                    raise TypeError(type_error_msg)
            else:
                raise TypeError(type_error_msg)
            if isinstance(tpl[1], string_types):
                names = [tpl[1]]
            elif isinstance(tpl[1], list):
                if all(isinstance(elem, string_types) for elem in tpl[1]):
                    names = tpl[1]
                else:
                    raise TypeError(type_error_msg)
            else:
                raise TypeError(type_error_msg)
            return self.__getitem(indices, names)

        if isinstance(key, Number):
            item = self.__getitem([key], self.variables)
        elif isinstance(key, string_types):
            item = self.__getitem(indices, [key])
        elif isinstance(key, list):
            item = getitem_list(key)
        elif isinstance(key, slice):
            item = getitem_list(list(range(key.start, key.stop, key.step or 1)))
        elif isinstance(key, tuple) and len(key) == 2:
            item = getitem_tuple(key)
        else:
            raise TypeError(type_error_msg)
        return item



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
API for uncertainty management
==============================

This API (Application Programming Interface) is dedicated to
uncertainty management. Current functions are:

- :meth:`~gemseo.uncertainty.api.get_available_distributions`,
- :meth:`~gemseo.uncertainty.api.create_distribution`,
- :meth:`~gemseo.uncertainty.api.create_statistics`.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

standard_library.install_aliases()

# pylint: disable=import-outside-toplevel


def get_available_distributions():
    """Get the available distributions."""
    from gemseo.uncertainty.distributions.factory import DistributionFactory

    factor = DistributionFactory()
    return factor.distributions


def create_distribution(variable, distribution_type, dimension=1, **options):
    """Create a distribution

    :param str variable: variable name.
    :param str distribution_type: distribution type.
    :param int dimension: variable dimension.
    :param options: distribution options.
    """
    from gemseo.uncertainty.distributions.factory import DistributionFactory

    factor = DistributionFactory()
    return factor.create(
        distribution_type, variable=variable, dimension=dimension, **options
    )


def create_statistics(
    dataset,
    variables_names=None,
    tested_distributions=None,
    fitting_criterion="BIC",
    selection_criterion="best",
    level=0.05,
    name=None,
):
    """Constructor

    :param Dataset dataset: dataset
    :param list(str) variables_names: list of variables names
        or list of variables names. If None, the method considers
        all variables from loaded dataset. Default: None.
    :param list(str) tested_distributions: list of candidate distributions
        names for parametric statistics. If None, considers empirical
        statistics. Default: None.
    :param str fitting_criterion: goodness-of-fit criterion
        for parametric statistics. Default: 'BIC'.
    :param float level: risk of committing a Type 1 error,
        that is an incorrect rejection of a true null hypothesis,
        for criteria based on test hypothesis in the case of
        parametric statistics. Default: 0.05.
    :param str selection_criterion: selection criterion
        for parametric statistics. Default: 'best'
    :param str name: name of the object.
        If None, use the concatenation of class and dataset names.
        Default: None.
    """
    from gemseo.uncertainty.statistics.empirical import EmpiricalStatistics as EmpStats
    from gemseo.uncertainty.statistics.parametric import (
        ParametricStatistics as ParamStats,
    )

    if tested_distributions is None:
        statistical_analysis = EmpStats(dataset, variables_names, name)
    else:
        statistical_analysis = ParamStats(
            dataset,
            tested_distributions,
            variables_names,
            fitting_criterion,
            level,
            selection_criterion,
            name,
        )
    return statistical_analysis



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Machine learning algorithm factory
==================================

This module contains a factory to instantiate a :class:`.MLAlgo` from its class
name. The class can be internal to |g| or located in an external module whose
path is provided to the constructor. It also provides a list of available
machine learning algorithm types and allows you to test if a machine learning
algorithm type is available.
"""
from __future__ import absolute_import, division, unicode_literals

import pickle
from os.path import join

from future import standard_library

from gemseo.core.factory import Factory
from gemseo.mlearning.core.ml_algo import MLAlgo

standard_library.install_aliases()


from gemseo import LOGGER


class MLAlgoFactory(object):
    """This factory instantiates a :class:`.MLAlgo`
    from its class name. The class can be internal to |g| or located in an
    external module whose path is provided to the constructor.
    """

    def __init__(self):
        """Initializes the factory: scans the directories to search for
        subclasses of MLAlgo. Searches in "GEMSEO_PATH" and gemseo.mlearning.
        """
        self.factory = Factory(MLAlgo, ("gemseo.mlearning",))

    def create(self, ml_algo, **options):
        """Create machine learning algorithm.

        :param str ml_algo: name of the machine learning algorithm (its
            classname).
        :param options: machine learning algorithm options.
        :return: MLAlgo
        """
        return self.factory.create(ml_algo, **options)

    @property
    def models(self):
        """Lists the available classes.

        :returns: list of class names.
        :rtype: list(str)
        """
        return self.factory.classes

    def is_available(self, ml_algo):
        """Checks the availability of a cache.

        :param str ml_algo:  name of the machine learning algorithm (its class
            name).
        :returns: True if the machine learning algorithm is available.
        :rtype: bool
        """
        return self.factory.is_available(ml_algo)

    def load(self, directory):
        """Load a machine learning algorithm from the disk.

        :param str directory: directory name.
        """
        with open(join(str(directory), MLAlgo.FILENAME), "rb") as handle:
            objects = pickle.load(handle)
        model = self.factory.create(
            objects.pop("algo_name"),
            data=objects.pop("data"),
            **objects.pop("parameters")
        )
        for key, value in objects.items():
            setattr(model, key, value)
        model.load_algo(directory)
        return model



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Generate a N2 and XDSM from an Excel description of the MDO problem
*******************************************************************
"""
from __future__ import absolute_import, division, unicode_literals

from ast import literal_eval

from future import standard_library
from pandas import read_excel
from six import string_types

from gemseo import LOGGER
from gemseo.api import (
    create_design_space,
    create_scenario,
    generate_n2_plot,
    get_available_formulations,
)
from gemseo.core.coupling_structure import MDOCouplingStructure
from gemseo.core.discipline import MDODiscipline

standard_library.install_aliases()


class XLSStudyParser(object):
    """
    Parse the input excel files that describe the GEMSEO study.

    The Excel file must contain one sheet per discipline.
    The name of the sheet is the discipline name.
    The sheet has at least two columns, one for inputs and one for outputs,
    with the following format:

    +--------+---------+----------------+--------------------+----------------+
    | Inputs | Outputs |Design variables| Objective function |  Constraints   |
    +========+=========+================+====================+================+
    |  in1   |  out1   |      in1       |       out1         |     out2       |
    +--------+---------+----------------+--------------------+----------------+
    |  in2   |  out2   |                |                    |                |
    +--------+---------+----------------+--------------------+----------------+

    Empty lines are ignored.
    All Objective functions and constraints must be outputs of a discipline,
    not necessarily the one of the current sheet
    All Design variables must be inputs of a discipline, not necessarily
    the one of the current sheet

    """

    SCENARIO_PREFIX = "Scenario"
    DISCIPLINE = "Discipline"
    DISCIPLINES = "Disciplines"
    OBJECTIVE_FUNCTION = "Objective function"
    CONSTRAINTS = "Constraints"
    DESIGN_VARIABLES = "Design variables"
    FORMULATION = "Formulation"
    OPTIONS = "Options"
    OPTIONS_VALUES = "Options values"

    def __init__(self, xls_study_path):
        """
        Initializes the study from the excel specification

        :param xls_study_path: path to the excel file describing the study
        """

        self.xls_study_path = xls_study_path
        try:
            self.frames = read_excel(xls_study_path, sheet_name=None, engine="openpyxl")
        except IOError:
            LOGGER.error("Failed to open study file !")
            raise

        LOGGER.info("Detected the following disciplines: %s", list(self.frames.keys()))
        self.disciplines = {}
        self.scenarios = {}
        self.inputs = []
        self.outputs = []

        self._init_disciplines()
        self._get_opt_pb_descr()

        if not self.scenarios:
            raise ValueError("Found no scenario in the xls file !")

    def _init_disciplines(self):
        """ Initializes disciplines. """
        for disc_name, frame in self.frames.items():
            if disc_name.startswith(self.SCENARIO_PREFIX):
                continue
            LOGGER.info("Parsing discipline %s", disc_name)

            try:
                inputs = self._get_frame_series_values(frame, "Inputs")
            except ValueError:
                raise ValueError(
                    "Discipline "
                    + str(disc_name)
                    + "'s sheet must have an Inputs column !"
                )
            self.inputs += inputs
            try:
                outputs = self._get_frame_series_values(frame, "Outputs")
            except ValueError:
                raise ValueError(
                    "Discipline "
                    + str(disc_name)
                    + "'s sheet must have an Outputs column !"
                )

            if not outputs:
                raise ValueError("Discipline " + str(disc_name) + " has no Outputs")
            self.outputs += outputs
            disc = MDODiscipline(disc_name)
            disc.input_grammar.initialize_from_data_names(inputs)
            disc.output_grammar.initialize_from_data_names(outputs)
            LOGGER.info("Inputs : %s", inputs)
            LOGGER.info("Outputs : %s", outputs)

            self.disciplines[disc_name] = disc

        self.inputs = set(self.inputs)
        self.outputs = set(self.outputs)

    @staticmethod
    def _get_frame_series_values(frame, series_name, return_none=False):
        """
        Gets the data list of a named column
        Removes empty data
        :param frame: the pandas frame of the sheet
        :param series_name: name of the series
        :param return_none: if the series does not exists, returns None
             instead of raising a ValueError
        """
        series = frame.get(series_name)
        if series is None:
            if return_none:
                return None
            raise ValueError("The sheet has no series named " + str(series_name))
        # Remove empty data
        # pylint: disable=comparison-with-itself
        return list([val for val in series.tolist() if val == val])

    def _get_opt_pb_descr(self):
        """
        Initilalize the objective function, constraints and design_variables
        """
        self.scenarios = {}

        for frame_name, frame in self.frames.items():
            if not frame_name.startswith(self.SCENARIO_PREFIX):
                continue
            LOGGER.info("Detected scenario in sheet: %s", frame_name)

            try:
                disciplines = self._get_frame_series_values(frame, self.DISCIPLINES)
            except ValueError:
                raise ValueError(
                    "Scenario "
                    + str(frame_name)
                    + " has no "
                    + self.DISCIPLINES
                    + " column !"
                )
            try:
                design_variables = self._get_frame_series_values(
                    frame, self.DESIGN_VARIABLES
                )
            except ValueError:
                raise ValueError(
                    "Scenario "
                    + str(frame_name)
                    + " has no "
                    + self.DESIGN_VARIABLES
                    + " column !"
                )
            try:
                objectives = self._get_frame_series_values(
                    frame, self.OBJECTIVE_FUNCTION
                )
            except ValueError:
                raise ValueError(
                    "Scenario "
                    + str(frame_name)
                    + " has no "
                    + self.OBJECTIVE_FUNCTION
                    + " column !"
                )
            try:
                constraints = self._get_frame_series_values(frame, self.CONSTRAINTS)
            except ValueError:
                raise ValueError(
                    "Scenario "
                    + str(frame_name)
                    + " has no "
                    + self.CONSTRAINTS
                    + " column !"
                )

            try:
                formulation = self._get_frame_series_values(frame, self.FORMULATION)
            except ValueError:
                raise ValueError(
                    "Scenario "
                    + str(frame_name)
                    + " has no "
                    + self.FORMULATION
                    + " column !"
                )

            options = self._get_frame_series_values(frame, self.OPTIONS, True)
            options_values = self._get_frame_series_values(
                frame, self.OPTIONS_VALUES, True
            )

            if len(formulation) != 1:
                raise ValueError(
                    "Scenario "
                    + str(frame_name)
                    + " must have 1 "
                    + self.FORMULATION
                    + " value !"
                )

            if options is not None:
                if len(options) != len(options_values):
                    raise ValueError(
                        "Options "
                        + str(options)
                        + " and Options values "
                        + str(options_values)
                        + " must have the same length!"
                    )

            formulation = formulation[0]

            scn = {}
            scn[self.DISCIPLINES] = disciplines
            scn[self.OBJECTIVE_FUNCTION] = objectives
            scn[self.CONSTRAINTS] = constraints
            scn[self.DESIGN_VARIABLES] = design_variables
            scn[self.FORMULATION] = formulation
            scn[self.OPTIONS] = options
            scn[self.OPTIONS_VALUES] = options_values

            self.scenarios[frame_name] = scn

        for name, desc in self.scenarios.items():
            self._check_opt_pb(
                desc[self.OBJECTIVE_FUNCTION],
                desc[self.CONSTRAINTS],
                desc[self.DISCIPLINES],
                desc[self.DESIGN_VARIABLES],
                desc[self.FORMULATION],
                name,
            )

    def _check_opt_pb(
        self,
        objectives,
        constraints,
        disciplines,
        design_variables,
        formulation,
        scn_name,
    ):
        """
        Checks the optimization problem consistency.
        Raises errors if needed

        :param objectives: list of objectives
        :param constraints: list of constraints
        :param disciplines: list of MDODisciplines
        :param design_variables: list of design varaibles
        :param formulation : mdo formulation name
        :param scn_name: name of the scenario

        """
        LOGGER.info("New scenario: %s", scn_name)
        LOGGER.info("Objectives: %s", objectives)
        LOGGER.info("Disciplines: %s", disciplines)
        LOGGER.info("Constraints: %s", constraints)
        LOGGER.info("Design variables: %s", design_variables)
        LOGGER.info("Formulation: %s", formulation)

        missing = set(design_variables) - self.inputs
        if missing:
            raise ValueError(
                scn_name + " : some design variables are "
                "not the inputs of any discipline :" + str(list(missing))
            )

        missing = (
            set(disciplines)
            - set(list(self.disciplines.keys()))
            - set(list(self.scenarios))
        )
        if missing:
            raise ValueError(
                scn_name + " : some disciplines dont exist :" + str(list(missing))
            )

        missing = set(constraints) - self.outputs
        if missing:
            raise ValueError(
                scn_name + " : some constraints are not "
                "the outputs of any discipline :" + str(list(missing))
            )

        missing = set(objectives) - self.outputs
        if missing:
            raise ValueError(
                scn_name + " : some objectives are not "
                "the outputs of any discipline :" + str(list(missing))
            )
        if not objectives:
            raise ValueError(scn_name + " : no objectives are defined!")

        if formulation not in get_available_formulations():
            raise ValueError(
                "Unknown formulation "
                + str(formulation)
                + " Use one of "
                + str(get_available_formulations())
            )


class StudyAnalysis(object):
    """
    Generate a N2 (equivalent to the Design Structure Matrix) diagram,
    showing the couplings between discipline and
    XDSM, (Extended Design Structure Matrix), showing the MDO process,
    from a Excel specification of the inputs, outputs, design variables,
    objectives and constraints.

    The input excel files contains one sheet per discipline.
    The name of the sheet is the discipline name.
    The sheet has at least two columns, one for inputs and one for outputs,
    with the following format:


    +--------+---------+
    | Inputs | Outputs |
    +========+=========+
    |  in1   |  out1   |
    +--------+---------+
    |  in2   |  out2   |
    +--------+---------+

    [Disc1]

    Empty lines are ignored.


    The scenarios (at least one, or multiple for distributed formulations)
    must appear in a Excel sheet name starting by "Scenario".

    The sheet has the following columns, with some constraints :
    All of them are mandatory, even if empty for the Constraints
    The order may be any
    1 and only 1 formulation must be declared
    At least 1 objective must be provided, and 1 design variable

    +----------------+--------------------+----------------+----------------+----------------+----------------+----------------+
    |Design variables| Objective function |  Constraints   |  Disciplines   |  Formulation   |  Options       | Options values |
    +================+====================+================+================+================+================+================+
    |      in1       |       out1         |     out2       |     Disc1      |     MDF        |  tolerance     |     0.1        |
    +----------------+--------------------+----------------+----------------+----------------+----------------+----------------+
    |                |                    |                |     Disc2      |                |                |                |
    +----------------+--------------------+----------------+----------------+----------------+----------------+----------------+

    [Scenario1]

    All Objective functions and constraints must be outputs of a discipline,
    not necessarily the one of the current sheet.
    All Design variables must be inputs of a discipline, not necessarily
    the one of the current sheet.

    The Options and Options values columns are used to pass
    the formulation options

    To use multi level MDO formulations, create multiple scenarios,
    and add the name of the sub scenarios
    in the list of disciplines of the main (system) scenario.

    An arbitrary number of levels can be generated this way
    (three, four levels etc formulations).
    """

    AVAILABLE_DISTRIBUTED_FORMULATIONS = ("BiLevel", "BLISS98B")

    def __init__(self, xls_study_path):
        """
        Initializes the study from the excel specification

        :param xls_study_path: path to the excel file describing the study
        """

        self.xls_study_path = xls_study_path
        self.study = XLSStudyParser(self.xls_study_path)
        self.disciplines_descr = self.study.disciplines
        self.scenarios_descr = self.study.scenarios
        self.disciplines = {}
        self.scenarios = {}
        self.main_scenario = None
        self._create_scenarios()

    def generate_n2(
        self,
        file_path="n2.pdf",
        show_data_names=True,
        save=True,
        show=False,
        figsize=(15, 10),
    ):
        """
        Generate a N2 plot for the disciplines list.

        :param file_path: File path of the figure.
        :type file_path: str
        :param show_data_names: If true, the names of the
            coupling data is shown
            otherwise, circles are drawn, which size depend on the
            number of coupling names.
        :type show_data_names: bool
        :param save: If True, saved the figure to file_path.
        :type save: bool
        :param show: If True, shows the plot.
        :type show: bool
        :param figsize: Size of the figure.
        :type figsize: tuple(float)
        """
        generate_n2_plot(
            list(self.disciplines.values()),
            file_path,
            show_data_names,
            save,
            show,
            figsize,
        )

    @staticmethod
    def _create_scenario(disciplines, scenario_descr):
        """
        Create a MDO scenario

        :param disciplines: list of MDODisciplines
        :param scenario_descr: description dict of the scenario
        :returns: the MDOScenario
        """
        coupl_struct = MDOCouplingStructure(disciplines)
        couplings = coupl_struct.get_all_couplings()
        design_space = create_design_space()
        scn_dv = scenario_descr[XLSStudyParser.DESIGN_VARIABLES]
        for var in set(scn_dv) | set(couplings):
            design_space.add_variable(var, size=1)

        options = scenario_descr[XLSStudyParser.OPTIONS]
        options_dict = {}
        if options is not None:
            options_values = scenario_descr[XLSStudyParser.OPTIONS_VALUES]
            for opt, val in zip(options, options_values):
                if isinstance(val, string_types):
                    try:
                        val = literal_eval(val)
                    except ValueError as err:
                        LOGGER.error(err)
                        raise ValueError(
                            "Failed to parse option " + str(opt) + " value :" + str(val)
                        )
                else:
                    pass
                options_dict[opt] = val

        scenario = create_scenario(
            disciplines,
            scenario_descr[XLSStudyParser.FORMULATION],
            scenario_descr[XLSStudyParser.OBJECTIVE_FUNCTION],
            design_space,
            **options_dict
        )
        for cstr in scenario_descr[XLSStudyParser.CONSTRAINTS]:
            scenario.add_constraint(cstr)
        return scenario

    def _get_disciplines_instances(self, scn):
        """
        Returns instances of the disciplines of the scenario,
        or None if not all available
        """
        discs = []
        for disc_name in scn[XLSStudyParser.DISCIPLINES]:
            disc_inst = self.disciplines_descr.get(disc_name)
            if disc_inst is None:  # not a discipline, so maybe a scenario
                disc_inst = self.scenarios.get(disc_name)
                if disc_inst is None:
                    return None
            discs.append(disc_inst)
        return discs

    def _create_scenarios(self):
        """
        Create the main scenario, eventually including sub scenarios
        """
        n_scn = len(self.scenarios_descr)
        i = 0

        while len(self.scenarios) != n_scn and i <= n_scn:
            i += 1
            for name, scn in self.scenarios_descr.items():
                discs = self._get_disciplines_instances(scn)
                if discs is not None:  # All depdendencies resolved
                    for disc in discs:
                        if not disc.is_scenario():
                            self.disciplines[disc.name] = disc

                    scenario = self._create_scenario(discs, scn)
                    self.scenarios[name] = scenario
                    # The last scenario created is the one
                    # with the most dependencies
                    # so the main one
                    self.main_scenario = scenario
        # At each while iteration at least 1 scenario must be resolved
        # otherwise this means there is a cross dependency between
        # scenarios
        if len(self.scenarios) != n_scn:
            raise ValueError(
                "Scenarios dependencies cannot be resolved,"
                " check for cycling dependencies "
                "between scenarios!"
            )

    def generate_xdsm(self, output_dir, latex_output=False, open_browser=False):
        """
        Creates an xdsm.json file from the current scenario.

        :param output_dir: the directory where XDSM html files are generated
        :param latex_output: build .tex, .tikz and .pdf file
        :returns: the MDOScenario, that contains the DesignSpace, the
            formulation, but the disciplines have only correct
            input and output grammars but no _run methods so that can't be executed
        """
        LOGGER.info("Generated the following Scenario:")
        self.main_scenario.log_me()
        self.main_scenario.formulation.opt_problem.log_me()
        self.main_scenario.xdsmize(
            outdir=output_dir, latex_output=latex_output, open_browser=open_browser
        )
        return self.main_scenario



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Random forest regression
========================

The random forest regression uses averaging methods on an ensemble
of decision trees.

Dependence
----------
The regression model relies on the RandomForestRegressor class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.ensemble.RandomForestRegressor.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from sklearn.ensemble import RandomForestRegressor as SKLRandForest

from gemseo.mlearning.regression.regression import MLRegressionAlgo

standard_library.install_aliases()


from gemseo import LOGGER


class RandomForestRegressor(MLRegressionAlgo):
    """ Random forest regression """

    LIBRARY = "scikit-learn"
    ABBR = "RandomForestRegressor"

    def __init__(
        self,
        data,
        transformer=None,
        input_names=None,
        output_names=None,
        n_estimators=100,
        **parameters
    ):
        """Constructor.

        :param data: learning dataset.
        :type data: Dataset
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param n_estimators: number of trees in the forest.
        :type n_estimators: int
        :param parameters: other keyword arguments for the sklearn algo.
        """
        super(RandomForestRegressor, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            n_estimators=n_estimators,
            **parameters
        )
        self.algo = SKLRandForest(n_estimators=n_estimators, **parameters)

    def _fit(self, input_data, output_data):
        """Fit the regression model.

        :param ndarray input_data: input data (2D)
        :param ndarray output_data: output data (2D)
        """
        # SKLearn RandomForestReressor does not like output
        # shape (n_samples, 1), prefers (n_samples,).
        # The shape (n_samples, n_outputs) with n_outputs >= 2 is fine.
        if output_data.shape[1] == 1:
            output_data = output_data[:, 0]
        self.algo.fit(input_data, output_data)

    def _predict(self, input_data):
        """Predict output for given input data.

        :param ndarray input_data: input data (2D).
        :return: output prediction (2D).
        :rtype: ndarray
        """
        output_data = self.algo.predict(input_data)

        # n_outputs=1 => output_shape=(n_samples,). Convert to (n_samples, 1).
        if len(output_data.shape) == 1:
            output_data = output_data[:, None]

        return output_data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Compute and display a Pareto Front
**********************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import range, str
from itertools import combinations

import matplotlib.pyplot as plt
from future import standard_library
from numpy import all as np_all
from numpy import any as np_any
from numpy import full, vstack

standard_library.install_aliases()


def select_pareto_optimal(obj_values):
    """
    Compute the Pareto front
    Search for all non dominated points, ie there exists j such that
    there is no lower value for obj_values[:,j] that does not degrade
    at least one other objective  obj_values[:,i]

    :param obj_values: objective function array, of size (n_samples, n_objs)
    :returns pareto_optimal: vector of booleans of size n_samples, True if
        Pareto optimal
    """
    pareto_optimal = full(obj_values.shape[0], True)

    def any_ax1_all(arr):
        return np_all(np_any(arr, axis=1))

    for i, obj in enumerate(obj_values):
        before_are_worse = any_ax1_all(obj_values[:i] > obj)
        after_are_worse = any_ax1_all(obj_values[i + 1 :] > obj)
        pareto_optimal[i] = before_are_worse and after_are_worse
    return pareto_optimal


def plot_pareto_bi_obj(axe, obj_values, pareto_optimal, obj_names, all_pareto=None):
    """
    Plot a 2D Pareto front

    :param axe: matplotlib axe on which to be plotted
    :param obj_values: objective function array, of size (n_samples, n_objs)
    :param pareto_optimal: vector of booleans of size n_samples,
        True if Pareto optimal.
    :param obj_names: names of the objectives
    :param all_pareto: indices of points that are pareto optimal wrt all
        criteria.
    """
    axe.scatter(obj_values[:, 0], obj_values[:, 1], color="b")

    if all_pareto is not None:
        axe.scatter(obj_values[all_pareto, 0], obj_values[all_pareto, 1], color="g")

    axe.scatter(obj_values[pareto_optimal, 0], obj_values[pareto_optimal, 1], color="r")

    axe.set_xlabel(obj_names[0])
    axe.set_ylabel(obj_names[1])


def generate_pareto_plots(obj_values, obj_names, figsize=(10, 10)):
    """
    Plot a 2D Pareto front
    :param  obj_values: objective function array, of size (n_samples, n_objs)
    :param obj_names: names of the objectives
    :param figsize: matplotlib figure size in inches
    """
    n_obj = obj_values.shape[1]
    if len(obj_names) != n_obj:
        raise ValueError(
            "Inconsistent objective values size and objective names!"
            + str(n_obj)
            + " !="
            + str(len(obj_names))
        )

    fig, axes = plt.subplots(n_obj - 1, n_obj - 1, figsize=figsize)

    fig.suptitle("Pareto front")

    pareto_opt_all = select_pareto_optimal(obj_values)
    # 0 vs 1   0 vs 2    0 vs 3
    #          1 vs 2    1 vs 3
    #                    2 vs 3

    # i j+1     i=0...nobj-1
    #           j=1....nobj

    for i, j in combinations(range(n_obj), 2):  # no duplication, j!=j
        if i == n_obj - 1:
            continue

        obj_loc = vstack((obj_values[:, i], obj_values[:, j])).T

        pareto_opt_loc = select_pareto_optimal(obj_loc)
        if n_obj == 2:
            axe = axes
        else:
            axe = axes[i, j - 1]
        plot_pareto_bi_obj(
            axe, obj_loc, pareto_opt_loc, [obj_names[i], obj_names[j]], pareto_opt_all
        )
        if i != j - 1:
            axes[j - 1, i].remove()

    return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Supervised machine learning algorithm
=====================================

Supervised machine learning is a task of learning relationships
between input and output variables based on an input-output dataset. One
usually distinguishes between to types of supervised machine learning
algorithms, based on the nature of the outputs. For a continuous output
variable, a regression is performed, while for a discrete output variable,
a classification is performed.

Given a set of input variables
:math:`x \\in \\mathbb{R}^{n_{\\text{samples}}\\times n_{\\text{inputs}}}` and
a set of output variables
:math:`y\\in \\mathbb{K}^{n_{\\text{samples}}\\times n_{\\text{outputs}}}`,
where :math:`n_{\\text{inputs}}` is the dimension of the input variable,
:math:`n_{\\text{outputs}}` is the dimension of the output variable,
:math:`n_{\\text{samples}}` is the number of training samples and
:math:`\\mathbb{K}` is either :math:`\\mathbb{R}` or :math:`\\mathbb{N}` for
regression and classification tasks respectively, a supervised learning
algorithm seeks to find a function
:math:`f: \\mathbb{R}^{n_{\\text{inputs}}} \\to
\\mathbb{K}^{n_{\\text{outputs}}}` such that :math:`y=f(x)`.

In addition, we often want to impose some additional constraints on the
function :math:`f`, mainly to ensure that it has a generalization capacity
beyond the training data, i.e. it is able to correctly predict output values of
new input values. This is called regularization. Assuming :math:`f` is
parametrized by a set of parameters :math:`\\theta`, and denoting
:math:`f_\\theta` the parametrized function, one typically seeks to minimize
a function of the form

.. math::

    \\mu(y, f_\\theta(x)) + \\Omega(\\theta),

where :math:`\\mu` is a distance-like measure, typically a mean squared error
or a cross entropy in the case of a regression, or a probability to be
maximized in the case of a classification, and :math:`\\Omega` is a
regularization term that limits the parameters from overfitting, typically some
norm of its argument.

The :mod:`~gemseo.mlearning.core.supervised` module implements this concept
through the :class:`.MLSupervisedAlgo` class based on a :class:`.Dataset`.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import atleast_2d

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import MLAlgo
from gemseo.mlearning.transform.dimension_reduction.dimension_reduction import (
    DimensionReduction,
)
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()


class MLSupervisedAlgo(MLAlgo):
    """Supervised machine learning algorithm.

    Inheriting classes should overload the :meth:`!MLSupervisedAlgo._fit` and
    :meth:`!MLSupervisedAlgo._predict` methods.
    """

    ABBR = "MLSupervisedAlgo"

    def __init__(
        self, data, transformer=None, input_names=None, output_names=None, **parameters
    ):
        """Constructor.

        :param Dataset data: learning dataset.
        :param transformer: transformation strategy for data groups.
            If None, do not scale data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param parameters: algorithm parameters.
        """
        super(MLSupervisedAlgo, self).__init__(
            data, transformer=transformer, **parameters
        )
        self.input_names = input_names or data.get_names(data.INPUT_GROUP)
        self.output_names = output_names or data.get_names(data.OUTPUT_GROUP)

    class DataFormatters(MLAlgo.DataFormatters):
        """Decorators for supervised algorithms. """

        @staticmethod
        def _array_to_dict(data_array, data_names, data_sizes):
            """Convert an array into a dict

            :param data_array: the array
            :param data_names: list of names (keys of the resulting dict)
            :param data_sizes: dict of (name, size)
            :returns: a dict
            :rtype: dict
            """
            current_position = 0
            array_dict = {}
            for name in data_names:
                array_dict[name] = data_array[
                    ..., current_position : current_position + data_sizes[name]
                ]
                current_position += data_sizes[name]
            return array_dict

        @classmethod
        def format_dict(cls, predict):
            """If input_data is passed as a dictionary, then convert it to
            ndarray, and convert output_data to dictionary. Else, do nothing.

            :param predict: Method whose input_data and output_data are to be
                formatted.
            """

            def wrapper(self, input_data, *args, **kwargs):
                as_dict = isinstance(input_data, dict)
                if as_dict:
                    input_data = DataConversion.dict_to_array(
                        input_data, self.input_names
                    )
                output_data = predict(self, input_data, *args, **kwargs)
                if as_dict:
                    varsizes = self.learning_set.sizes
                    output_data = cls._array_to_dict(
                        output_data, self.output_names, varsizes
                    )
                return output_data

            return wrapper

        @classmethod
        def format_samples(cls, predict):
            """If input_data has shape (n_inputs,), reshape input_data to
            (1, n_inputs), and then reshape output data from (1, n_outputs)
            to (n_outputs,).
            If input_data has shape (n_samples, n_inputs), then do nothing.

            :param predict: Method whose input_data and output_data are to be
                formatted.
            """

            def wrapper(self, input_data, *args, **kwargs):
                """Format data before and after applying predictor. """
                single_sample = input_data.ndim == 1
                input_data = atleast_2d(input_data)
                output_data = predict(self, input_data, *args, **kwargs)
                if single_sample:
                    output_data = output_data[0]
                return output_data

            return wrapper

        @classmethod
        def format_transform(cls, transform_inputs=True, transform_outputs=True):
            """Apply transform to inputs, and inverse transform to outputs.

            :param bool format_inputs: Indicates whether to transform inputs.
            :param bool format_outputs: Indicates whether to transform outputs.
            """

            def format_transform_(predict):
                """Apply transform to inputs, and inverse transform to
                outputs.

                :param predict: Method whose input_data and output_data are to
                    be formatted.
                """

                def wrapper(self, input_data, *args, **kwargs):
                    """Wrapped version of predict function. """
                    inputs = self.learning_set.INPUT_GROUP
                    if transform_inputs and inputs in self.transformer:
                        input_data = self.transformer[inputs].transform(input_data)
                    output_data = predict(self, input_data, *args, **kwargs)
                    outputs = self.learning_set.OUTPUT_GROUP
                    if transform_outputs and outputs in self.transformer:
                        output_data = self.transformer[outputs].inverse_transform(
                            output_data
                        )
                    return output_data

                return wrapper

            return format_transform_

        @classmethod
        def format_input_output(cls, predict):
            """Format dict, samples and transform successively.

            :param predict: Method whose input_data and output_data are to be
                formatted.
            """

            @cls.format_dict
            @cls.format_samples
            @cls.format_transform()
            def wrapper(self, input_data, *args, **kwargs):
                return predict(self, input_data, *args, **kwargs)

            return wrapper

    def learn(self, samples=None):
        """Train machine learning algorithm on learning set, possibly filtered
        using the given parameters.

        :param list(int) samples: indices of training samples.
        """
        input_grp = self.learning_set.INPUT_GROUP
        output_grp = self.learning_set.OUTPUT_GROUP
        input_data = self.learning_set.get_data_by_names(self.input_names, False)
        output_data = self.learning_set.get_data_by_names(self.output_names, False)

        if samples is not None:
            input_data = input_data[samples]
            output_data = output_data[samples]

        if input_grp in self.transformer:
            input_data = self.transformer[input_grp].fit_transform(input_data)

        if output_grp in self.transformer:
            output_data = self.transformer[output_grp].fit_transform(output_data)

        self._fit(input_data, output_data)
        self._trained = True

    def _fit(self, input_data, output_data):
        """Fit input-output relationship from data learning.

        :param ndarray input_data: input data (2D).
        :param ndarray output_data: output data (2D).
        """
        raise NotImplementedError

    @DataFormatters.format_input_output
    def predict(self, input_data):
        """Predict output data from input data.

        :param input_data: input data (n_inputs,) or (n_samples, n_inputs).
        :type input_data: dict(ndarray) or ndarray
        :return: predicted output data (n_outputs,) or (n_samples, n_outputs).
        :rtype: dict(ndarray) or ndarray(int)
        """
        return self._predict(input_data)

    def _predict(self, input_data):
        """Predict output data from input data.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :return: output data (n_samples, n_outputs).
        :rtype: ndarray(int)
        """
        raise NotImplementedError

    def _get_raw_shapes(self):
        """Get raw input and output shapes.

        The raw dimensions are the shapes of input and output variables after
        applying transformers.

        return: raw input shape, raw output shape
        rtype: tuple(int)
        """
        reduce_inputs = Dataset.INPUT_GROUP in self.transformer and isinstance(
            self.transformer[Dataset.INPUT_GROUP], DimensionReduction
        )
        if reduce_inputs:
            input_shape = self.transformer[Dataset.INPUT_GROUP].n_components
        else:
            input_shape = self.input_shape

        reduce_outputs = Dataset.OUTPUT_GROUP in self.transformer and isinstance(
            self.transformer[Dataset.OUTPUT_GROUP], DimensionReduction
        )
        if reduce_outputs:
            output_shape = self.transformer[Dataset.OUTPUT_GROUP].n_components
        else:
            output_shape = self.output_shape

        return input_shape, output_shape

    @property
    def input_shape(self):
        """ Dimension of input variables before applying transformers. """
        sizes = [self.learning_set.sizes[name] for name in self.input_names]
        return sum(sizes)

    @property
    def output_shape(self):
        """ Dimension of output variables before applying transformers. """
        sizes = [self.learning_set.sizes[name] for name in self.output_names]
        return sum(sizes)

    def _get_objects_to_save(self):
        """Get objects to save.

        :return: objects to save.
        :rtype: dict
        """
        objects = super(MLSupervisedAlgo, self)._get_objects_to_save()
        objects["input_names"] = self.input_names
        objects["output_names"] = self.output_names
        return objects



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A Gauss Seidel algorithm for solving MDAs
*****************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import super

from future import standard_library

from gemseo.core.chain import MDOChain
from gemseo.core.discipline import MDODiscipline
from gemseo.mda.mda import MDA

standard_library.install_aliases()


class MDAGaussSeidel(MDA):
    """
    Perform a MDA analysis using a Gauss-Seidel algorithm,
    an iterative technique to solve the linear system:

    .. math::

       Ax = b

    by decomposing the matrix :math:`A`
    into the sum of a lower triangular matrix :math:`L_*`
    and a strictly upper triangular matrix :math:`U`.

    The new iterate is given by:

    .. math::

       x_{k+1} = L_*^{-1}(b-Ux_k)
    """

    def __init__(
        self,
        disciplines,
        name=None,
        max_mda_iter=10,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
        tolerance=1e-6,
        linear_solver_tolerance=1e-12,
        warm_start=False,
        use_lu_fact=False,
        norm0=None,
    ):
        """
        Constructor

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param max_mda_iter: maximum number of iterations
        :type max_mda_iter: int
        :param name: the name of the chain
        :type name: str
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param linear_solver_tolerance: Tolerance of the linear solver
            in the adjoint equation
        :type linear_solver_tolerance: float
        :param warm_start: if True, the second iteration and ongoing
            start from the previous coupling solution
        :type warm_start: bool
        :param use_lu_fact: if True, when using adjoint/forward
            differenciation, store a LU factorization of the matrix
            to solve faster multiple RHS problem
        :type use_lu_fact: bool
        :param norm0: reference value of the norm of the residual to compute
            the decrease stop criteria.
            Iterations stops when norm(residual)/norm0<tolerance
        :type norm0: float
        """
        self.chain = MDOChain(disciplines, grammar_type=grammar_type)
        super(MDAGaussSeidel, self).__init__(
            disciplines,
            max_mda_iter=max_mda_iter,
            name=name,
            grammar_type=grammar_type,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            warm_start=warm_start,
            use_lu_fact=use_lu_fact,
        )
        self._initialize_grammars()
        self._set_default_inputs()
        self._compute_input_couplings()

    def reset_statuses_for_run(self):
        """Sets all the statuses to PENDING"""
        super(MDAGaussSeidel, self).reset_statuses_for_run()
        self.chain.reset_statuses_for_run()

    def _initialize_grammars(self):
        """Defines all inputs and outputs of the chain"""
        # self.chain.initialize_grammars()
        self.input_grammar.update_from(self.chain.input_grammar)
        self.output_grammar.update_from(self.chain.output_grammar)

    def _run(self):
        """Runs the disciplines in a sequential way until the difference
        between outputs is under tolerance

        :returns: the local data
        """
        if self.warm_start:
            self._couplings_warm_start()
        # execute the disciplines
        current_couplings = self._current_input_couplings()
        self.chain.reset_statuses_for_run()
        self.local_data.update(self.chain.execute(self.local_data))
        new_couplings = self._current_input_couplings()

        # store initial residual
        current_iter = 1
        self._compute_residual(
            current_couplings, new_couplings, current_iter, first=True
        )
        current_couplings = new_couplings

        while not self._termination(current_iter):
            self.chain.reset_statuses_for_run()
            self.local_data.update(self.chain.execute(self.local_data))
            new_couplings = self._current_input_couplings()

            # store current residual
            current_iter += 1
            self._compute_residual(current_couplings, new_couplings, current_iter)
            current_couplings = new_couplings



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2014 Jon Wolverton github.com/wolverdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
A JSON schema generation tool
*****************************
"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from builtins import dict, str, super
from collections import defaultdict
import json
import re

from future import standard_library
from six import string_types


standard_library.install_aliases()


JS_TYPES = {dict: 'object',
            list: 'array',
            str: 'string',
            type('a'): 'string',
            type(u'a'): 'string',
            int: 'integer',
            float: 'number',
            bool: 'boolean',
            type(None): 'null'}

for typ in string_types:
    JS_TYPES[typ] = "string"


class SchemaDefaultDict(defaultdict):
    """A defaultdict implementation that is serializable with pickle
    and that returns an empty Schema by default
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor
        """
        super(SchemaDefaultDict, self).__init__()
        self.options = {}

    def __missing__(self, key):
        value = Schema(**self.options)
        self[key] = value
        return value


class Schema(object):
    """Basic schema generator class. Schema objects can be loaded up
    with existing schemas and objects before being serialized.
    """

    def __init__(self, merge_arrays=True, additional_items=True,
                 additional_props=True, match_props=[],
                 exclude_at_merge=None):
        """Builds a schema generator object.

        * `merge_arrays` (default `True`): Assume all array items share
          the same schema (as they should). The alternate behavior is to
          create a different schema for each item in every array.
        * `additional_items` (default 'True'): If True, allow
          tuple-validated arrays to be followed by unvalidated items.
          If False, generate an "additionalItems": False element so that
          array items not specified in the schema cause a ValidationError.
          Not used with list-validated (merged) arrays.
        * `additional_props` (default 'True'): If True, allow objects
          to include unvalidated properties.  If False, generate an
          "additionalProperties": False element so that properties not
          specified in the schema cause a ValidationError.
        * `match_props` (default '[]'): List of regular expressions to
          compare with property keys.  Properties with matching
          keys share the same "patternProperties" schema.

        """

        self._options = {
            'merge_arrays': merge_arrays,
            'additional_items': additional_items,
            'additional_props': additional_props,
            'match_props': match_props,
        }
        self._type = set()
        self._description = None
        self._required = None
        self._exclude_at_merge = exclude_at_merge
        if self._exclude_at_merge is None:
            self._exclude_at_merge = []
# Original implementation
#         self._properties = DictWithDefaultSchema(**self._options)
#         self._patternProperties = DictWithDefaultSchema(**self._options)
        self._patternProperties = SchemaDefaultDict()
        self._patternProperties.options = self._options
        self._properties = SchemaDefaultDict()
        self._properties.options = self._options
        self._items = []
        self._other = {}

    def add_schema(self, schema):
        """Merges in an existing schema.

        :param schema: required
        :param an: existing JSON Schema to merge
        """

        # serialize instances of Schema before parsing
        if isinstance(schema, Schema):
            schema = schema.to_dict()

        # parse properties and add them individually
        for prop, val in list(schema.items()):
            if prop == 'type':
                self._add_type(val)
            elif prop == 'required':
                self._add_required(val)
            elif prop in ['properties', 'patternProperties']:
                self._add_properties(prop, val, 'add_schema')
            elif prop == 'items':
                self._add_items(val, 'add_schema')
            elif prop == "description":
                self.add_description(val)
            elif prop not in self._other:
                self._other[prop] = val
            elif self._other[prop] != val and prop not in\
                    self._exclude_at_merge:
                e = prop + ': ' + str(self._other[prop]) + ' ^= ' + str(val)
                raise SchemaError('schema incompatible -- ' + e)

        # make sure the 'required' key gets set regardless
        if 'required' not in schema:
            self._add_required([])

        # return self for easy method chaining
        return self

    def add_object(self, obj, descr=None):
        """Modify the schema to accomodate an object.

        :param obj: required
        :param a: JSON object to use in generate the schema
        """

        if isinstance(obj, dict):
            self._generate_object(obj, descr)
        elif isinstance(obj, list):
            self._generate_array(obj, descr)
        else:
            self._generate_basic(obj, descr)

        # return self for easy method chaining
        return self

    def to_dict(self):
        """Convert the current schema to a `dict`.

        """
        # start with existing fields
        schema = dict(self._other)

        if 'additionalItems' in schema:
            if schema['additionalItems'] == True or not isinstance(
                    self._items, list):
                del(schema['additionalItems'])
        if 'additionalProperties' in schema and schema[
                'additionalProperties'] == True:
            del(schema['additionalProperties'])

        # unpack the type field
        if self._type:
            schema['type'] = self._get_type()

        if self._description:
            schema["description"] = self._get_description()

        # call recursively on subschemas if object or array
        if 'object' in self._type:
            props = self._get_properties()
            # include unnecessary but valid "properties": {}
            if props[0] or not props[1]:
                schema['properties'] = props[0]
            if props[1]:
                schema['patternProperties'] = props[1]
            if self._required:
                schema['required'] = self._get_required()
        if 'array' in self._type:
            items = self._get_items()
            if items or isinstance(items, dict):
                schema['items'] = items
        return schema

    def to_json(self, *args, **kwargs):
        """Convert the current schema directly to serialized JSON.

        :param args:
        :param kwargs:
        """
        return json.dumps(self.to_dict(), *args, **kwargs)

    # private methods

    # getters

    def _get_type(self):
        """ """
        schema_type = self._type | set()  # get a copy

        # remove any redundant integer type
        if 'integer' in schema_type and 'number' in schema_type:
            schema_type.remove('integer')

        # unwrap if only one item, else convert to array
        if len(schema_type) == 1:
            (schema_type,) = schema_type
        else:
            schema_type = sorted(schema_type)

        return schema_type

    def _get_description(self):
        return self._description

    def _get_required(self):
        """ """
        return sorted(self._required) if self._required else []

    def _get_properties(self):
        """


        """
        properties, patprops = {}, {}
        for prop, subschema in list(self._properties.items()):
            properties[prop] = subschema.to_dict()
        for prop, subschema in list(self._patternProperties.items()):
            patprops[prop] = subschema.to_dict()
        return ((properties, patprops))

    def _get_items(self):
        """
        Lists items

        """
        if isinstance(self._items, list):
            return [subschema.to_dict() for subschema in self._items]
        else:
            return self._items.to_dict()

    # setters

    def _add_type(self, val_type):
        """

        :param val_type:
        """
        if isinstance(val_type, string_types):
            self._type.add(val_type)
        else:
            self._type |= set(val_type)

    def _add_required(self, required):
        """

        :param required:

        """
        if self._required is None:
            # if not already set, set to this
            self._required = set(required)
        else:
            # use intersection to limit to properties present in both
            self._required.update(set(required))

    def _add_properties(self, ptype, properties, func):
        """

        :param ptype: param properties:
        :param func:
        :param properties:
        """
        # recursively modify subschemas
        pattern = self._options['match_props']
        if pattern and not ptype:
            self._add_properties_merge(pattern, properties, func)
        else:
            self._add_properties_sep(ptype or 'properties', properties, func)

    def _add_properties_merge(self, pattern, properties, func):
        """

        :param pattern: param properties:
        :param func:
        :param properties:
        """
        err = None
        for prop, val in list(properties.items()):
            match = []
            for pat in pattern:
                m = re.search(pat, prop)
                if m:
                    getattr(self._patternProperties[pat], func)(val)
                    match.append(pat)
                    self._required -= set((prop,))
            if len(match) > 1:
                err = (prop, match)
            elif len(match) < 1:
                getattr(self._properties[prop], func)(val)
        if err:
            raise SchemaError('patternProperties multiple match: ' +
                              err[0] + ' ~= ' + ', '.join(err[1]))

    def _add_properties_sep(self, ptype, properties, func):
        """

        :param ptype: param properties:
        :param func:
        :param properties:
        """
        pdict = self._properties if ptype == 'properties' else self._patternProperties
        for prop, val in list(properties.items()):
            getattr(pdict[prop], func)(val)

    def _add_items(self, items, func):
        """

        :param items: param func:
        :param func:
        """
        if self._options['merge_arrays']:
            self._add_items_merge(items, func)
        else:
            self._add_items_sep(items, func)

    def add_description(self, descr):
        """

        :param descr: tue description
        """
        self._description = descr

    def _add_items_merge(self, items, func):
        """

        :param items: param func:
        :param func:
        """
        if not self._items:
            self._items = Schema(**self._options)
        method = getattr(self._items, func)
        if isinstance(items, list):
            for item in items:
                method(item)
        else:
            method(items)

    def _add_items_sep(self, items, func):
        """

        :param items: param func:
        :param func:
        """
        for item in items:
            subschema = Schema(**self._options)
            getattr(subschema, func)(item)
            self._items.append(subschema)

    def _add_additionalItems(self):
        """ """
        self._other['additionalItems'] = self._options['additional_items']

    def _add_additionalProperties(self):
        """ """
        self._other['additionalProperties'] = self._options['additional_props']

    # generate from object

    def _generate_object(self, obj, descr=None):
        """

        :param obj:
        """
        self._add_type('object')
        self._add_required(list(obj.keys()))
        self._add_properties(None, obj, 'add_object')
        self._add_additionalProperties()
        if descr is not None:
            self._add_properties(None, descr, 'add_description')

    def _generate_array(self, array, descr=None):
        """

        :param array:
        """
        self._add_type('array')
        self._add_items(array, 'add_object')
        self._add_additionalItems()
        if descr is not None:
            self.add_description(descr)

    def _generate_basic(self, val, descr=None):
        """

        :param val:
        """
        val_type = JS_TYPES[type(val)]
        if val_type != "null":
            self._add_type(val_type)
        if descr is not None:
            self.add_description(descr)


class SchemaError(Exception):
    """ """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Dataset plot factory
====================

The module :mod:`~gemseo.post.dataset.factory` contains
the :class:`DatasetPlotFactory` class which is a factory
to instantiate a :class:`.DatasetPlot` from its class name.
The class can be internal to |g| or located in an external module whose path
is provided to the constructor. It also provides a list of available cache
types and allows you to test if a cache type is available.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.core.factory import Factory
from gemseo.post.dataset.dataset_plot import DatasetPlot

standard_library.install_aliases()


from gemseo import LOGGER


class DatasetPlotFactory(object):
    """This factory instantiates a :class:`.DatasetPlot` from its class name.
    The class can be internal to |g| or located in an external module
    whose path is provided to the constructor.
    """

    def __init__(self):
        """
        Initializes the factory: scans the directories to search for
        subclasses of DatasetPlot.
        Searches in "GEMSEO_PATH" and gemseo.mlearning.p_datasets
        """
        self.factory = Factory(DatasetPlot, ("gemseo.post.dataset",))

    def create(self, plot_name, **options):
        """
        Create a plot for dataset

        :param str plot_name: name of the plot for dataset
            (its classname)
        :param options: additional options specific
        :return: dataset plot
        """
        return self.factory.create(plot_name, **options)

    @property
    def plots(self):
        """
        Lists the available plots for datasets.

        :returns: the list of plots for datasets.
        :rtype: list(str)
        """
        return self.factory.classes

    def is_available(self, plot_name):
        """
        Checks the availability of a plot for dataset.

        :param str plot_name:  name of the plot for dataset
            (its class name).
        :returns: True if the plot for dataset is available.
        :rtype: bool
        """
        return self.factory.is_available(plot_name)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""
Polynomial chaos expansion
==========================

The polynomial chaos expansion (PCE) discipline expresses the model output
as a weighted sum of polynomial functions which are orthonormal
in the stochastic input space spanned by the random input variables:

.. math::

    Y = w_0 + w_1\phi_1(X) + w_2\phi_2(X) + ... + w_K\phi_K(X)

where :math:`\phi_i(x)=\phi_{\tau_1(i),1}(x_1)\times\ldots\times
\phi_{\tau_d(i),d}(x_d)`.

Enumerating strategy
--------------------

The choice of the function :math:`\tau=(\tau_1,\ldots,\tau_d)` is an
enumerating strategy and :math:`\tau_j(i)` is the polynomial degree of
:math:`\psi_{\tau_j(i),j}`.

Distributions
-------------

PCE are stochastic models whose inputs are random variables
and are often used to deal with uncertainty quantification problems.

If :math:`X_j` is a Gaussian random variable,
:math:`(\psi_{ij})_{i\geq 0}` is the Legendre basis.
If :math:`X_j` is an uniform random variable,
:math:`(\psi_{ij})_{i\geq 0}` is the Hermite basis.

When the problem is deterministic, we can still use PCE
under the assumptions that
the random variables are independent uniform random variables.
Then, the orthonormal basis function is the Hermite basis.

Degree
------

The degree :math:`P` of a PCE is defined
in such a way that :math:`\text{degree}(\psi_i)=\sum_{j=1}^d\tau_j(i)\leq P`.

Estimation
----------

The coefficients :math:`(w_1, w_2, ..., w_K)` and the intercept
:math:`w_0` are estimated either by least square regression,
sparse least square regression or quadrature.

Dependence
----------
The PCE model relies on the FunctionalChaosAlgorithm class
of the `openturns library <http://openturns.github.io/openturns/1.9/user_manual
/response_surface/_generated/openturns.FunctionalChaosAlgorithm.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import all as np_all
from numpy import array, concatenate, isin, zeros, zeros_like
from openturns import (
    LARS,
    AdaptiveStieltjesAlgorithm,
    CleaningStrategy,
    ComposedDistribution,
    CorrectedLeaveOneOut,
    EnumerateFunction,
    FixedStrategy,
    Function,
    FunctionalChaosAlgorithm,
    FunctionalChaosSobolIndices,
    GaussProductExperiment,
    IntegrationStrategy,
    LeastSquaresMetaModelSelectionFactory,
    LeastSquaresStrategy,
    MarginalTransformationEvaluation,
    OrthogonalBasis,
    OrthogonalProductPolynomialFactory,
    Point,
    StandardDistributionPolynomialFactory,
)

from gemseo.mlearning.regression.regression import MLRegressionAlgo

standard_library.install_aliases()


from gemseo import LOGGER


class PCERegression(MLRegressionAlgo):
    """ Polynomial chaos expansion. """

    LIBRARY = "openturns"
    ABBR = "PCE"
    LS_STRATEGY = "LS"
    QUAD_STRATEGY = "Quad"
    SPARSE_STRATEGY = "SparseLS"
    AVAILABLE_STRATEGIES = [LS_STRATEGY, QUAD_STRATEGY, SPARSE_STRATEGY]

    def __init__(
        self,
        data,
        probability_space,
        discipline=None,
        transformer=None,
        input_names=None,
        output_names=None,
        strategy=LS_STRATEGY,
        degree=2,
        n_quad=None,
        stieltjes=True,
        sparse_param=None,
    ):
        """Constructor.

        :param data: learning dataset
        :type data: Dataset
        :param probability_space: probability space.
        :type probability_space: ParameterSpace
        :param discipline: discipline to evaluate if strategy='Quad' and
            data is empty.
        :type discipline: MDODiscipline
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param strategy: strategy to compute the parameters of the PCE,
            either 'LS', 'Quad' or 'SparseLS'. Default: 'LS'.
        :type strategy: str
        :param degree: polynomial degree of the PCE
        :type degree: int
        :param n_quad: number of quadrature points
        :type n_quad: int
        :param stieltjes: stieltjes
        :type stieltjes: bool
        :param sparse_param: Parameters for the Sparse Cleaning Truncation
            Strategy and/or hyperbolic truncation of the initial basis:

            - **max_considered_terms** (int) -- Maximum Considered Terms,
            - **most_significant** (int), Most Significant number to retain,
            - **significance_factor** (float), Significance Factor,
            - **hyper_factor** (float), factor for hyperbolic truncation
              strategy.
        """
        prob_space = probability_space
        super(PCERegression, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            probability_space=prob_space,
            strategy=strategy,
            degree=degree,
            n_quad=n_quad,
            stieltjes=stieltjes,
            sparse_param=sparse_param,
        )
        self.prob_space = prob_space
        self.discipline = discipline
        if transformer is not None and transformer != {}:
            raise ValueError("PCERegression does not support transformers.")
        try:
            if data:
                u_names = set(prob_space.variables_names)
                assert set(self.input_names) == u_names
            else:
                self.input_names = prob_space.variables_names
        except Exception:
            raise ValueError(
                "Data inputs names are %s "
                "while probability distributions are defined "
                "%s." % (self.input_names, list(prob_space.marginals.keys()))
            )

        self.distributions = prob_space.marginals
        self.sparse_param = sparse_param or {}
        self.input_dim = sum([dist.dimension for _, dist in self.distributions.items()])
        self.strategy = strategy
        if self.strategy not in self.AVAILABLE_STRATEGIES:
            strategies = " ".join(self.AVAILABLE_STRATEGIES)
            raise ValueError(
                "{} is a wrong strategy."
                " Available ones are: {}".format(self.strategy, strategies)
            )
        self.stieltjes = stieltjes
        self.degree = degree
        self.n_quad = n_quad
        self.ot_distributions = [
            self.distributions[name].distribution for name in self.input_names
        ]
        self.dist = ComposedDistribution(self.ot_distributions)
        hyper_factor = self.sparse_param.get("hyper_factor", 1.0)
        self.enumerate_function = EnumerateFunction(self.input_dim, hyper_factor)
        self.basis = self._get_basis()
        self.n_basis = self._get_basis_size()
        self.trunc_strategy = self._get_trunc_strategy()
        if self.strategy == self.QUAD_STRATEGY:
            (
                self.sample,
                self.weights,
                self.proj_strategy,
            ) = self._get_quadrature_points()
        else:
            self.sample = None
            self.weights = None

    def _fit(self, input_data, output_data):
        """Fit the regression model.

        :param ndarray input_data: input data (2D)
        :param ndarray output_data: output data (2D)
        """
        self.proj_strategy = self._get_proj_strategy(input_data, output_data)
        weights = self._get_weights(input_data)
        self.algo = self._build_pce(input_data, weights, output_data)

    def _predict(self, input_data):
        """Predict output.

        :param ndarray input_data: input data (2D).
        :return: output prediction (2D).
        :rtype: ndarray
        """
        return array(self.algo.getMetaModel()(input_data))

    def compute_sobol(self):
        """Compute first and total Sobol' indices

        :returns: first and total SObol' indices
        :rtype: list, list
        """
        sensitivity_analysis = FunctionalChaosSobolIndices(self.algo)
        LOGGER.info(str(sensitivity_analysis))
        dimension = len(self.input_names)
        first_order = [sensitivity_analysis.getSobolIndex(i) for i in range(dimension)]
        total_order = [
            sensitivity_analysis.getSobolTotalIndex(i) for i in range(dimension)
        ]
        return first_order, total_order

    def _build_pce(self, x_learn, weights, y_learn):
        """Build PCE"""
        pce_algo = FunctionalChaosAlgorithm(
            x_learn, weights, y_learn, self.dist, self.trunc_strategy
        )
        pce_algo.setProjectionStrategy(self.proj_strategy)
        pce_algo.run()
        return pce_algo.getResult()

    def _get_basis(self):
        """Get basis function for PCE construction"""
        if self.stieltjes:
            # Tend to result in performance issue
            basis = OrthogonalProductPolynomialFactory(
                [
                    StandardDistributionPolynomialFactory(
                        AdaptiveStieltjesAlgorithm(marginal)
                    )
                    for marginal in self.ot_distributions
                ],
                self.enumerate_function,
            )
        else:
            basis = OrthogonalProductPolynomialFactory(
                [
                    StandardDistributionPolynomialFactory(margin)
                    for margin in self.ot_distributions
                ],
                self.enumerate_function,
            )
        return basis

    def _get_basis_size(self):
        """Get basis size for PCE construction"""
        return self.enumerate_function.getStrataCumulatedCardinal(self.degree)

    def _get_quadrature_points(self):
        """Get quadrature points for PCE construction"""
        measure = self.basis.getMeasure()
        if self.n_quad is not None:
            degree_by_dim = int(self.n_quad ** (1.0 / self.input_dim))
        else:
            degree_by_dim = self.degree + 1
        degrees = [degree_by_dim] * self.input_dim

        proj_strategy = IntegrationStrategy(GaussProductExperiment(measure, degrees))
        sample, weights = proj_strategy.getExperiment().generateWithWeights()

        if not self.stieltjes:
            transformation = Function(
                MarginalTransformationEvaluation(
                    [measure.getMarginal(i) for i in range(self.input_dim)],
                    self.ot_distributions,
                    False,
                )
            )
            sample = transformation(sample)
        sample = array(sample)
        if not self.learning_set:
            inputs_names = self.prob_space.variables_names
            in_grp = self.learning_set.INPUT_GROUP
            self.learning_set.set_from_array(
                sample,
                self.prob_space.variables_names,
                self.prob_space.variables_sizes,
                {name: in_grp for name in inputs_names},
            )
        if self.discipline is not None:
            n_samples = len(self.learning_set)
            outputs = {name: [] for name in self.discipline.get_output_data_names()}
            for data in self.learning_set:
                output_data = self.discipline.execute(data)
                for name in self.discipline.get_output_data_names():
                    outputs[name] += list(output_data[name])
            for name in self.discipline.get_output_data_names():
                outputs[name] = array(outputs[name]).reshape((n_samples, -1))
            outputs_names = list(self.discipline.get_output_data_names())
            outputs = [outputs[name] for name in outputs_names]
            data = concatenate(outputs, axis=1)
            out_grp = self.learning_set.OUTPUT_GROUP
            sizes = {
                name: len(self.discipline.local_data[name]) for name in outputs_names
            }
            self.learning_set.add_group(
                out_grp, data, outputs_names, sizes, cache_as_input=False
            )
            self.output_names = outputs_names

        return sample, weights, proj_strategy

    def _get_trunc_strategy(self):
        """Get truncation strategy for PCE construction"""
        if self.strategy in [self.LS_STRATEGY, self.QUAD_STRATEGY]:
            trunc_strategy = FixedStrategy(self.basis, self.n_basis)
        elif self.strategy == self.SPARSE_STRATEGY:
            sparse_param = self.sparse_param
            sparse_param = {} if sparse_param is None else sparse_param
            max_considered_terms = sparse_param.get("max_considered_terms", 120)
            most_significant = sparse_param.get("most_significant", 30)
            significance_factor = sparse_param.get("significance_factor", 1e-3)

            trunc_strategy = CleaningStrategy(
                OrthogonalBasis(self.basis),
                max_considered_terms,
                most_significant,
                significance_factor,
                True,
            )
        return trunc_strategy

    def _get_proj_strategy(self, x_learn, y_learn):
        """Get projection strategy for PCE construction

        :param x_learn: input data
        :type x_learn: array
        :param y_learn: output data
        :type y_learn: array
        """
        if self.strategy == self.QUAD_STRATEGY:
            proj_strategy = self.proj_strategy
        elif self.strategy == self.LS_STRATEGY:
            proj_strategy = LeastSquaresStrategy(x_learn, y_learn)
        else:
            app = LeastSquaresMetaModelSelectionFactory(LARS(), CorrectedLeaveOneOut())
            proj_strategy = LeastSquaresStrategy(x_learn, y_learn, app)
        return proj_strategy

    def _get_ls_weights(self):
        """Get LS weights for PCE construction"""
        _, weights = self.proj_strategy.getExperiment().generateWithWeights()
        return weights

    def _get_quad_weights(self, x_learn):
        """Get quadrature weights for PCE construction"""
        sample = zeros_like(self.sample)
        common_len = len(x_learn)
        sample[:common_len] = x_learn
        sample[0] = self.sample[0]
        sample_arg = np_all(isin(sample, self.sample), axis=1)
        new_weights = array(self.weights)[sample_arg]
        return new_weights

    def _get_weights(self, x_learn):
        """Get weights for PCE construction)"""
        if self.strategy == self.QUAD_STRATEGY:
            weights = self._get_quad_weights(x_learn)
        else:
            weights = self._get_ls_weights()
        return weights

    def _predict_jacobian(self, input_data):
        """Predict Jacobian of the regression model for the given input data.

        :param ndarray input_data: input_data (2D).
        :return: Jacobian matrices (3D, one for each sample).
        :rtype: ndarray
        """
        input_shape, output_shape = self._get_raw_shapes()
        gradient = self.algo.getMetaModel().gradient
        jac = zeros((input_data.shape[0], int(output_shape), int(input_shape)))
        for index, data in enumerate(input_data):
            jac[index] = array(gradient(Point(data))).T
        return jac



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Francois Gallard
"""
Run a DOE from a file containing samples values
***********************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from future import standard_library
from numpy import atleast_2d, loadtxt

from gemseo.algos.doe.doe_lib import DOELibrary

standard_library.install_aliases()


from gemseo import LOGGER


class CustomDOE(DOELibrary):

    """Class used for creation of DOE samples provided by user
    This samples are provided as file in text or csv format.
    """

    ALGO_LIST = ["CustomDOE"]
    DELIMITER_KEYWORD = "delimiter"
    SKIPROWS_KEYWORD = "skiprows"
    DOE_FILE = "doe_file"

    def __init__(self):
        """
        Constructor, initializes the DOE samples
        For this class of DOE library, samples are provided as file
        in text or csv format. csv file format is assume to have a header
        whereas text file (extension .txt) has not.
        """
        super(CustomDOE, self).__init__()
        self.file_dv_names_list = None

        desc = {}
        desc["CustomDOE"] = (
            "The **CustomDOE** class is used for creation"
            " of DOE samples provided by user. This samples"
            " are provided as file in text or csv format."
        )
        for algo in self.ALGO_LIST:
            self.lib_dict[algo] = {
                DOELibrary.LIB: self.__class__.__name__,
                DOELibrary.INTERNAL_NAME: algo,
                DOELibrary.DESCRIPTION: desc[algo],
            }

    def _get_options(
        self,
        doe_file,
        delimiter=",",  # pylint: disable=W0221
        comments="#",
        skiprows=0,
        eval_jac=False,
        n_processes=1,
        wait_time_between_samples=0.0,
        **kwargs
    ):
        """Sets the options

        :param doe_file: path and name of file
        :type doe_file: str
        :param delimiter: The string used to separate values.
        :type delimiter: str
        :param comments:  the characters or list of characters used to
            indicate the start of a comment
        :type comments: str
        :param skiprows: skip the first `skiprows` lines
        :type skiprows: int
        :param eval_jac: evaluate jacobian
        :type eval_jac: bool
        :param n_processes: number of processes
        :type n_processes: int
        :param wait_time_between_samples: waiting time between two samples
        :type wait_time_between_samples: float
        :param kwargs: additional arguments

        """
        wtbs = wait_time_between_samples
        return self._process_options(
            doe_file=doe_file,
            delimiter=delimiter,
            comments=comments,
            skiprows=skiprows,
            eval_jac=eval_jac,
            n_processes=n_processes,
            wait_time_between_samples=wtbs,
            **kwargs
        )

    def read_file(self, doe_file, delimiter=",", comments="#", skiprows=0):
        """Read a file containing a DOE

        :param doe_file: path and name of file
        :type doe_file: str
        :param delimiter: The string used to separate values.
        :type delimiter: str
        :param comments:  the characters or list of characters used to
            indicate the start of a comment
        :type comments: str
        :param skiprows: skip the first `skiprows` lines
        :type skiprows: int
        :returns: sample (an array of samples)
        :rtype: numpy array
        """
        try:
            samples = loadtxt(
                doe_file,
                comments=comments,
                delimiter=delimiter,
                skiprows=skiprows,
                unpack=False,
            )
            samples = atleast_2d(samples)
            if (
                samples.shape[1] != self.problem.dimension
                and self.problem.dimension == 1
            ):
                samples = samples.T
        except ValueError:
            LOGGER.error("Failed to load DOE input file : %s", str(doe_file))
            raise
        self.__check_input_dv_lenght(samples)

        # Normalize samples
        normalize_vect = self.problem.design_space.normalize_vect
        for i in range(samples.shape[0]):
            samples[i, :] = normalize_vect(samples[i, :])

        return samples

    def __check_input_dv_lenght(self, samples):
        """
        Check that file contains all variables given as design variable
        at initialization
        """
        dim = self.problem.dimension
        if samples.shape[1] != dim:
            raise ValueError(
                "Mismatch between problem design variables "
                + str(dim)
                + " and doe input file dimension : "
                + str(samples.shape[1])
            )

    def _generate_samples(self, **options):
        """
        Generates the list of x samples

        :param options: the options dict for the algorithm,
            see associated JSON file
        """

        samples = self.read_file(
            options[self.DOE_FILE],
            delimiter=options[self.DELIMITER_KEYWORD],
            skiprows=options[self.SKIPROWS_KEYWORD],
        )
        return samples



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Machine learning API
--------------------

The machine learning API provides methods for creating new and loading
existing machine learning models. It also provides methods for listing
available models and options.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.api import _get_schema

standard_library.install_aliases()

# pylint: disable=import-outside-toplevel


def get_mlearning_models():
    """Get available machine learning algorithms.

    See also
    --------
    import_mlearning_model
    create_mlearning_model
    get_mlearning_options
    import_mlearning_model
    """
    from gemseo.mlearning.core.factory import MLAlgoFactory

    factory = MLAlgoFactory()
    return factory.models


def get_regression_models():
    """Get available regression models.

    See also
    --------
    create_regression_model
    get_regression_options
    import_regression_model
    """
    from gemseo.mlearning.regression.factory import RegressionModelFactory

    factory = RegressionModelFactory()
    return factory.models


def get_classification_models():
    """Get available classification models.

    See also
    --------
    create_classification_model
    get_classification_options
    import_classification_model
    """
    from gemseo.mlearning.classification.factory import ClassificationModelFactory

    factory = ClassificationModelFactory()
    return factory.models


def get_clustering_models():
    """Get available clustering models.

    See also
    --------
    create_clustering_model
    get_clustering_options
    import_clustering_model
    """
    from gemseo.mlearning.cluster.factory import ClusteringModelFactory

    factory = ClusteringModelFactory()
    return factory.models


def create_mlearning_model(name, data, transformer=None, **parameters):
    """Create machine learning algorithm from a learning data set.

    :param str name: name of the machine learning algorithm.
    :param Dataset data: learning data set.
    :param dict(str) transformer: transformation strategy for data groups.
        If None, do not transform data. Default: None.
    :param parameters: machine learning algorithm parameters.

    See also
    --------
    get_mlearning_models
    get_mlearning_options
    import_mlearning_model
    """
    from gemseo.mlearning.core.factory import MLAlgoFactory

    factory = MLAlgoFactory()
    return factory.create(name, data=data, transformer=transformer, **parameters)


def create_regression_model(name, data, transformer=None, **parameters):
    """Create a regression model from a learning data set.

    :param str name: name of the regression model.
    :param Dataset data: learning data set.
    :param dict(str) transformer: transformation strategy for data groups.
        If None, do not transform data. Default: None.
    :param parameters: regression model parameters.

    See also
    --------
    get_regression_models
    get_regression_options
    import_regression_model
    """
    from gemseo.mlearning.regression.factory import RegressionModelFactory

    factory = RegressionModelFactory()
    return factory.create(name, data=data, transformer=transformer, **parameters)


def create_classification_model(name, data, transformer=None, **parameters):
    """Create a classification model from a learning data set.

    :param str name: name of the classification model.
    :param Dataset data: learning data set.
    :param dict(str) transformer: transformation strategy for data groups.
        If None, do not transform data. Default: None.
    :param parameters: classification model parameters.

    See also
    --------
    get_classification_models
    get_classification_options
    import_classification_model
    """
    from gemseo.mlearning.classification.factory import ClassificationModelFactory

    factory = ClassificationModelFactory()
    return factory.create(name, data=data, transformer=transformer, **parameters)


def create_clustering_model(name, data, transformer=None, **parameters):
    """Create a clustering model from a learning data set.

    :param str name: name of the clustering model.
    :param Dataset data: learning data set.
    :param dict(str) transformer: transformation strategy for data groups.
        If None, do not transform data. Default: None.
    :param parameters: clustering model parameters.

    See also
    --------
    get_clustering_models
    get_clustering_options
    import_clustering_model
    """
    from gemseo.mlearning.cluster.factory import ClusteringModelFactory

    factory = ClusteringModelFactory()
    return factory.create(name, data=data, transformer=transformer, **parameters)


def import_mlearning_model(directory):
    """Import a machine learning algorithm from a directory.

    :param str directory: directory name.

    See also
    --------
    create_mlearning_model
    get_mlearning_models
    get_mlearning_options
    """
    from gemseo.mlearning.core.factory import MLAlgoFactory

    factory = MLAlgoFactory()
    return factory.load(directory)


def import_regression_model(directory):
    """Import a regression model from a directory.

    :param str directory: directory name.

    See also
    --------
    create_regression_model
    get_regression_models
    get_regression_options
    """
    from gemseo.mlearning.regression.factory import RegressionModelFactory

    factory = RegressionModelFactory()
    return factory.load(directory)


def import_classification_model(directory):
    """Import a classification model from a directory.

    :param str directory: directory name.

    See also
    --------
    create_classification_model
    get_classification_models
    get_classification_options
    """
    from gemseo.mlearning.classification.factory import ClassificationModelFactory

    factory = ClassificationModelFactory()
    return factory.load(directory)


def import_clustering_model(directory):
    """Import a clustering model from a directory.

    :param str directory: directory name.

    See also
    --------
    create_clustering_model
    get_clustering_models
    get_clustering_options
    """
    from gemseo.mlearning.cluster.factory import ClusteringModelFactory

    factory = ClusteringModelFactory()
    return factory.load(directory)


def get_mlearning_options(model_name, output_json=False, pretty_print=True):
    """
    Lists the available options for a machine learning algorithm.

    :param str model_name: Name of the machine learning algorithm.
    :param bool output_json: Apply json format for the schema.
    :param bool pretty_print: Print the schema in a pretty table.
    :returns: Option schema (string) of the machine learning algorithm.

    See also
    --------
    create_mlearning_model
    get_mlearning_models
    import_mlearning_model
    """
    from gemseo.mlearning.core.factory import MLAlgoFactory

    factory = MLAlgoFactory().factory
    grammar = factory.get_options_grammar(model_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_regression_options(model_name, output_json=False, pretty_print=True):
    """
    Lists the available options for a regression model.

    :param str model_name: Name of the regression model.
    :param bool output_json: Apply json format for the schema.
    :param bool pretty_print: Print the schema in a pretty table.
    :returns: Option schema (string) of the regression model.

    See also
    --------
    create_regression_model
    get_regression_models
    import_regression_model
    """
    from gemseo.mlearning.regression.factory import RegressionModelFactory

    factory = RegressionModelFactory().factory
    grammar = factory.get_options_grammar(model_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_classification_options(model_name, output_json=False, pretty_print=True):
    """
    Lists the available options for a classification model.

    :param str model_name: Name of the classification model.
    :param bool output_json: Apply json format for the schema.
    :param bool pretty_print: Print the schema in a pretty table.
    :returns: Option schema (string) of the classification model.

    See also
    --------
    create_classification_model
    get_classification_models
    import_classification_model
    """
    from gemseo.mlearning.classification.factory import ClassificationModelFactory

    factory = ClassificationModelFactory().factory
    grammar = factory.get_options_grammar(model_name)
    return _get_schema(grammar, output_json, pretty_print)


def get_clustering_options(model_name, output_json=False, pretty_print=True):
    """
    Lists the available options for clustering model.

    :param str model_name: Name of the clustering model.
    :param bool output_json: Apply json format for the schema.
    :param bool pretty_print: Print the schema in a pretty table.
    :returns: Option schema (string) of the clustering model.

    See also
    --------
    create_clustering_model
    get_clustering_models
    import_clustering_model
    """
    from gemseo.mlearning.cluster.factory import ClusteringModelFactory

    factory = ClusteringModelFactory().factory
    grammar = factory.get_options_grammar(model_name)
    return _get_schema(grammar, output_json, pretty_print)
